import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("example", "0002_taggeditem"),
    ]

    operations = [
        migrations.CreateModel(
            name="Company",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Project",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("topic", models.CharField(max_length=30)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AlterField(
            model_name="comment",
            name="entry",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="comments",
                to="example.Entry",
            ),
        ),
        migrations.CreateModel(
            name="ArtProject",
            fields=[
                (
                    "project_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="example.Project",
                    ),
                ),
                ("artist", models.CharField(max_length=30)),
            ],
            options={
                "abstract": False,
            },
            bases=("example.project",),
        ),
        migrations.CreateModel(
            name="ResearchProject",
            fields=[
                (
                    "project_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="example.Project",
                    ),
                ),
                ("supervisor", models.CharField(max_length=30)),
            ],
            options={
                "abstract": False,
            },
            bases=("example.project",),
        ),
        migrations.AddField(
            model_name="project",
            name="polymorphic_ctype",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="polymorphic_example.project_set+",
                to="contenttypes.ContentType",
            ),
        ),
        migrations.AddField(
            model_name="company",
            name="current_project",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="companies",
                to="example.Project",
            ),
        ),
        migrations.AddField(
            model_name="company",
            name="future_projects",
            field=models.ManyToManyField(to="example.Project"),
        ),
    ]



import pytest
from django.urls import reverse
from rest_framework import status

pytestmark = pytest.mark.django_db


def test_sparse_fieldset_valid_fields(client, entry):
    base_url = reverse("entry-list")
    response = client.get(base_url, data={"fields[entries]": "blog,headline"})
    assert response.status_code == status.HTTP_200_OK
    data = response.json()["data"]

    assert len(data) == 1
    entry = data[0]
    assert entry["attributes"].keys() == {"headline"}
    assert entry["relationships"].keys() == {"blog"}


@pytest.mark.parametrize(
    "fields_param", ["invalidfields[entries]", "fieldsinvalid[entries"]
)
def test_sparse_fieldset_invalid_fields_parameter(client, entry, fields_param):
    """
    Test that invalid fields query parameter is not processed by sparse fieldset.

    rest_framework_json_api.filters.QueryParameterValidationFilter takes care of error
    handling in such a case.
    """
    base_url = reverse("entry-list")
    response = client.get(base_url, data={"invalidfields[entries]": "blog,headline"})
    assert response.status_code == status.HTTP_200_OK
    data = response.json()["data"]

    assert len(data) == 1
    entry = data[0]
    assert entry["attributes"].keys() != {"headline"}
    assert entry["relationships"].keys() != {"blog"}



import pytest
from django.urls import path, reverse
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from rest_framework_json_api import serializers
from rest_framework_json_api.parsers import JSONParser
from rest_framework_json_api.relations import ResourceRelatedField
from rest_framework_json_api.renderers import JSONRenderer
from rest_framework_json_api.utils import format_link_segment
from rest_framework_json_api.views import ModelViewSet
from tests.models import BasicModel


class TestModelViewSet:
    @pytest.mark.parametrize(
        "format_links",
        [
            False,
            "dasherize",
            "camelize",
            "capitalize",
            "underscore",
        ],
    )
    def test_get_related_field_name_handles_formatted_link_segments(
        self, settings, format_links, rf
    ):
        settings.JSON_API_FORMAT_RELATED_LINKS = format_links

        # use field name which actually gets formatted
        related_model_field_name = "related_field_model"

        class RelatedFieldNameSerializer(serializers.ModelSerializer):
            related_model_field = ResourceRelatedField(queryset=BasicModel.objects)

            def __init__(self, *args, **kwargs):
                self.related_model_field.field_name = related_model_field_name
                super().__init(*args, **kwargs)

            class Meta:
                model = BasicModel

        class RelatedFieldNameView(ModelViewSet):
            serializer_class = RelatedFieldNameSerializer

        url_segment = format_link_segment(related_model_field_name)

        request = rf.get(f"/basic_models/1/{url_segment}")

        view = RelatedFieldNameView()
        view.setup(request, related_field=url_segment)

        assert view.get_related_field_name() == related_model_field_name


class TestAPIView:
    @pytest.mark.urls(__name__)
    def test_patch(self, client):
        data = {
            "data": {
                "id": 123,
                "type": "custom",
                "attributes": {"body": "hello"},
            }
        }

        url = reverse("custom")

        response = client.patch(url, data=data)
        result = response.json()

        assert result["data"]["id"] == str(123)
        assert result["data"]["type"] == "custom"
        assert result["data"]["attributes"]["body"] == "hello"


class CustomModel:
    def __init__(self, response_dict):
        for k, v in response_dict.items():
            setattr(self, k, v)

    @property
    def pk(self):
        return self.id if hasattr(self, "id") else None


class CustomModelSerializer(serializers.Serializer):
    body = serializers.CharField()
    id = serializers.IntegerField()


class CustomAPIView(APIView):
    parser_classes = [JSONParser]
    renderer_classes = [JSONRenderer]
    resource_name = "custom"

    def patch(self, request, *args, **kwargs):
        serializer = CustomModelSerializer(CustomModel(request.data))
        return Response(status=status.HTTP_200_OK, data=serializer.data)


urlpatterns = [
    path("custom", CustomAPIView.as_view(), name="custom"),
]



from rest_framework_json_api.relations import ResourceRelatedField
from rest_framework_json_api.serializers import ModelSerializer
from tests.models import (
    BasicModel,
    ForeignKeySource,
    ForeignKeyTarget,
    ManyToManySource,
    ManyToManyTarget,
)


class BasicModelSerializer(ModelSerializer):
    class Meta:
        fields = ("text",)
        model = BasicModel


class ForeignKeySourceSerializer(ModelSerializer):
    target = ResourceRelatedField(queryset=ForeignKeyTarget.objects)

    class Meta:
        model = ForeignKeySource
        fields = ("target",)


class ManyToManySourceSerializer(ModelSerializer):
    targets = ResourceRelatedField(many=True, queryset=ManyToManyTarget.objects)

    class Meta:
        model = ManyToManySource
        fields = ("targets",)


class ManyToManyTargetSerializer(ModelSerializer):
    class Meta:
        model = ManyToManyTarget


class ManyToManySourceReadOnlySerializer(ModelSerializer):
    targets = ResourceRelatedField(many=True, read_only=True)

    class Meta:
        model = ManyToManySource
        fields = ("targets",)



# Generated by Django 3.2.3 on 2021-05-26 03:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("example", "0008_labresults"),
    ]

    operations = [
        migrations.AddField(
            model_name="labresults",
            name="author",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="lab_results",
                to="example.author",
            ),
        ),
    ]



import rest_framework.exceptions as exceptions
import rest_framework.parsers
import rest_framework.renderers
from django_filters import rest_framework as filters
from rest_framework.filters import SearchFilter

import rest_framework_json_api.metadata
import rest_framework_json_api.parsers
import rest_framework_json_api.renderers
from rest_framework_json_api.django_filters import DjangoFilterBackend
from rest_framework_json_api.filters import (
    OrderingFilter,
    QueryParameterValidationFilter,
)
from rest_framework_json_api.pagination import JsonApiPageNumberPagination
from rest_framework_json_api.utils import format_drf_errors
from rest_framework_json_api.views import (
    ModelViewSet,
    ReadOnlyModelViewSet,
    RelationshipView,
)

from example.models import (
    Author,
    Blog,
    Comment,
    Company,
    Entry,
    LabResults,
    Project,
    ProjectType,
)
from example.serializers import (
    AuthorDetailSerializer,
    AuthorListSerializer,
    AuthorSerializer,
    BlogDRFSerializer,
    BlogSerializer,
    CommentSerializer,
    CompanySerializer,
    EntryDRFSerializers,
    EntrySerializer,
    LabResultsSerializer,
    ProjectSerializer,
    ProjectTypeSerializer,
)

HTTP_422_UNPROCESSABLE_ENTITY = 422


class BlogViewSet(ModelViewSet):
    queryset = Blog.objects.all()
    serializer_class = BlogSerializer

    def get_object(self):
        entry_pk = self.kwargs.get("entry_pk", None)
        if entry_pk is not None:
            return Entry.objects.get(id=entry_pk).blog

        return super().get_object()


class DRFBlogViewSet(ModelViewSet):
    queryset = Blog.objects.all()
    serializer_class = BlogDRFSerializer
    lookup_url_kwarg = "entry_pk"

    def get_object(self):
        entry_pk = self.kwargs.get(self.lookup_url_kwarg, None)
        if entry_pk is not None:
            return Entry.objects.get(id=entry_pk).blog

        return super().get_object()


class JsonApiViewSet(ModelViewSet):
    """
    This is an example on how to configure DRF-jsonapi from
    within a class. It allows using DRF-jsonapi alongside
    vanilla DRF API views.
    """

    parser_classes = [
        rest_framework_json_api.parsers.JSONParser,
        rest_framework.parsers.FormParser,
        rest_framework.parsers.MultiPartParser,
    ]
    renderer_classes = [
        rest_framework_json_api.renderers.JSONRenderer,
        rest_framework.renderers.BrowsableAPIRenderer,
    ]
    metadata_class = rest_framework_json_api.metadata.JSONAPIMetadata

    def handle_exception(self, exc):
        if isinstance(exc, exceptions.ValidationError):
            # some require that validation errors return 422 status
            # for example ember-data (isInvalid method on adapter)
            exc.status_code = HTTP_422_UNPROCESSABLE_ENTITY
        # exception handler can't be set on class so you have to
        # override the error response in this method
        response = super().handle_exception(exc)
        context = self.get_exception_handler_context()
        return format_drf_errors(response, context, exc)


class BlogCustomViewSet(JsonApiViewSet):
    queryset = Blog.objects.all()
    serializer_class = BlogSerializer


class EntryViewSet(ModelViewSet):
    queryset = Entry.objects.all()
    resource_name = "posts"

    def get_serializer_class(self):
        return EntrySerializer

    def get_object(self):
        # Handle featured
        entry_pk = self.kwargs.get("entry_pk", None)
        if entry_pk is not None:
            return Entry.objects.exclude(pk=entry_pk).first()

        return super().get_object()


class DRFEntryViewSet(ModelViewSet):
    queryset = Entry.objects.all()
    serializer_class = EntryDRFSerializers
    lookup_url_kwarg = "entry_pk"

    def get_object(self):
        # Handle featured
        entry_pk = self.kwargs.get(self.lookup_url_kwarg, None)
        if entry_pk is not None:
            return Entry.objects.exclude(pk=entry_pk).first()

        return super().get_object()


class NoPagination(JsonApiPageNumberPagination):
    page_size = None


class NonPaginatedEntryViewSet(EntryViewSet):
    pagination_class = NoPagination
    # override the default filter backends in order to test QueryParameterValidationFilter without
    # breaking older usage of non-standard query params like `page_size`.
    filter_backends = (
        QueryParameterValidationFilter,
        OrderingFilter,
        DjangoFilterBackend,
        SearchFilter,
    )
    ordering_fields = ("headline", "body_text", "blog__name", "blog__id")
    rels = (
        "exact",
        "iexact",
        "contains",
        "icontains",
        "gt",
        "gte",
        "lt",
        "lte",
        "in",
        "regex",
        "isnull",
    )
    filterset_fields = {
        "id": ("exact", "in"),
        "headline": rels,
        "body_text": rels,
        "blog__name": rels,
        "blog__tagline": rels,
    }
    search_fields = ("headline", "body_text", "blog__name", "blog__tagline")


class EntryFilter(filters.FilterSet):
    bname = filters.CharFilter(field_name="blog__name", lookup_expr="exact")

    authors__id = filters.ModelMultipleChoiceFilter(
        field_name="authors",
        to_field_name="id",
        conjoined=True,  # to "and" the ids
        queryset=Author.objects.all(),
    )

    class Meta:
        model = Entry
        fields = {
            "id": ("exact",),
            "headline": ("exact",),
            "body_text": ("exact",),
            "authors__id": ("in",),
        }


class FiltersetEntryViewSet(EntryViewSet):
    """
    like above but use filterset_class instead of filterset_fields
    """

    pagination_class = NoPagination
    filterset_fields = None
    filterset_class = EntryFilter
    filter_backends = (
        QueryParameterValidationFilter,
        DjangoFilterBackend,
    )


class NoFiltersetEntryViewSet(EntryViewSet):
    """
    like above but no filtersets
    """

    pagination_class = NoPagination
    filterset_fields = None
    filterset_class = None


class AuthorViewSet(ModelViewSet):
    queryset = Author.objects.all()

    def get_serializer_class(self):
        serializer_classes = {
            "list": AuthorListSerializer,
            "retrieve": AuthorDetailSerializer,
        }

        action = getattr(self, "action", "")
        return serializer_classes.get(action, AuthorSerializer)


class CommentViewSet(ModelViewSet):
    queryset = Comment.objects.all()
    serializer_class = CommentSerializer
    select_for_includes = {"writer": ["author__bio"]}
    prefetch_for_includes = {
        "__all__": [],
        "author": ["author__bio", "author__entries"],
    }

    def get_queryset(self, *args, **kwargs):
        queryset = super().get_queryset()

        entry_pk = self.kwargs.get("entry_pk", None)
        if entry_pk is not None:
            queryset = queryset.filter(entry_id=entry_pk)

        return queryset


class CompanyViewset(ModelViewSet):
    queryset = Company.objects.all()
    serializer_class = CompanySerializer


class ProjectViewset(ModelViewSet):
    queryset = Project.objects.all().order_by("pk")
    serializer_class = ProjectSerializer


class ProjectTypeViewset(ModelViewSet):
    queryset = ProjectType.objects.all()
    serializer_class = ProjectTypeSerializer


class EntryRelationshipView(RelationshipView):
    queryset = Entry.objects.all()


class BlogRelationshipView(RelationshipView):
    queryset = Blog.objects.all()


class CommentRelationshipView(RelationshipView):
    queryset = Comment.objects.all()


class AuthorRelationshipView(RelationshipView):
    queryset = Author.objects.all()
    self_link_view_name = "author-relationships"


class LabResultViewSet(ReadOnlyModelViewSet):
    queryset = LabResults.objects.all()
    serializer_class = LabResultsSerializer
    prefetch_for_includes = {
        "__all__": [],
        "author": ["author__bio", "author__entries"],
    }



import pytest
from rest_framework.test import APIClient

from tests.models import (
    BasicModel,
    ForeignKeyTarget,
    ManyToManySource,
    ManyToManyTarget,
)


@pytest.fixture(autouse=True)
def use_rest_framework_json_api_defaults(settings):
    """
    Enfroce default settings for tests modules.


    As for now example and tests modules share the same settings file
    some defaults which have been overwritten in the example app need
    to be overwritten. This way testing actually happens on default resp.
    each test defines what non default setting it wants to test.

    Once migration to tests module is finished and tests can have
    its own settings file, this fixture can be removed.
    """

    settings.JSON_API_FORMAT_FIELD_NAMES = False
    settings.JSON_API_FORMAT_TYPES = False
    settings.JSON_API_PLURALIZE_TYPES = False


@pytest.fixture
def model(db):
    return BasicModel.objects.create(text="Model")


@pytest.fixture
def foreign_key_target(db):
    return ForeignKeyTarget.objects.create(name="Target")


@pytest.fixture
def many_to_many_source(db, many_to_many_targets):
    source = ManyToManySource.objects.create(name="Source")
    source.targets.add(*many_to_many_targets)
    return source


@pytest.fixture
def many_to_many_targets(db):
    return [
        ManyToManyTarget.objects.create(name="Target1"),
        ManyToManyTarget.objects.create(name="Target2"),
    ]


@pytest.fixture
def client():
    return APIClient()



# Generated by Django 2.1.4 on 2018-12-28 07:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("example", "0005_auto_20180922_1508"),
    ]

    operations = [
        migrations.CreateModel(
            name="AuthorBioMetadata",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("body", models.TextField()),
                (
                    "bio",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metadata",
                        to="example.AuthorBio",
                    ),
                ),
            ],
            options={
                "ordering": ("id",),
            },
        ),
        migrations.AlterField(
            model_name="comment",
            name="author",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="comments",
                to="example.Author",
            ),
        ),
    ]



import inspect
import operator
import warnings
from collections import OrderedDict

import inflection
from django.conf import settings
from django.db.models import Manager
from django.db.models.fields.related_descriptors import (
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
)
from django.http import Http404
from django.utils import encoding
from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions, relations
from rest_framework.exceptions import APIException

from .settings import json_api_settings

# Generic relation descriptor from django.contrib.contenttypes.
if "django.contrib.contenttypes" not in settings.INSTALLED_APPS:  # pragma: no cover
    # Target application does not use contenttypes. Importing would cause errors.
    ReverseGenericManyToOneDescriptor = object()
else:
    from django.contrib.contenttypes.fields import ReverseGenericManyToOneDescriptor


def get_resource_name(context, expand_polymorphic_types=False):
    """
    Return the name of a resource.
    """
    from rest_framework_json_api.serializers import PolymorphicModelSerializer

    view = context.get("view")

    # Sanity check to make sure we have a view.
    if not view:
        return None

    # Check to see if there is a status code and return early
    # with the resource_name value of `errors`.
    try:
        code = str(view.response.status_code)
    except (AttributeError, ValueError):
        pass
    else:
        if code.startswith("4") or code.startswith("5"):
            return "errors"

    try:
        resource_name = getattr(view, "resource_name")
    except AttributeError:
        try:
            if "kwargs" in context and "related_field" in context["kwargs"]:
                serializer = view.get_related_serializer_class()
            else:
                serializer = view.get_serializer_class()
            if expand_polymorphic_types and issubclass(
                serializer, PolymorphicModelSerializer
            ):
                return serializer.get_polymorphic_types()
            else:
                return get_resource_type_from_serializer(serializer)
        except AttributeError:
            try:
                resource_name = get_resource_type_from_model(view.model)
            except AttributeError:
                resource_name = view.__class__.__name__

            if not isinstance(resource_name, str):
                # The resource name is not a string - return as is
                return resource_name

            # the name was calculated automatically from the view > pluralize and format
            resource_name = format_resource_type(resource_name)

    return resource_name


def get_serializer_fields(serializer):
    fields = None
    if hasattr(serializer, "child"):
        fields = getattr(serializer.child, "fields")
        meta = getattr(serializer.child, "Meta", None)
    if hasattr(serializer, "fields"):
        fields = getattr(serializer, "fields")
        meta = getattr(serializer, "Meta", None)

    if fields is not None:
        meta_fields = getattr(meta, "meta_fields", {})
        for field in meta_fields:
            try:
                fields.pop(field)
            except KeyError:
                pass
        return fields


def format_field_names(obj, format_type=None):
    """
    Takes a dict and returns it with formatted keys as set in `format_type`
    or `JSON_API_FORMAT_FIELD_NAMES`

    :format_type: Either 'dasherize', 'camelize', 'capitalize' or 'underscore'
    """
    if format_type is None:
        format_type = json_api_settings.FORMAT_FIELD_NAMES

    if isinstance(obj, dict):
        formatted = OrderedDict()
        for key, value in obj.items():
            key = format_value(key, format_type)
            formatted[key] = value
        return formatted

    return obj


def undo_format_field_names(obj):
    """
    Takes a dict and undo format field names to underscore which is the Python convention
    but only in case `JSON_API_FORMAT_FIELD_NAMES` is actually configured.
    """
    if json_api_settings.FORMAT_FIELD_NAMES:
        return format_field_names(obj, "underscore")

    return obj


def format_field_name(field_name):
    """
    Takes a field name and returns it with formatted keys as set in
    `JSON_API_FORMAT_FIELD_NAMES`
    """
    return format_value(field_name, json_api_settings.FORMAT_FIELD_NAMES)


def undo_format_field_name(field_name):
    """
    Takes a string and undos format field name to underscore which is the Python convention
    but only in case `JSON_API_FORMAT_FIELD_NAMES` is actually configured.
    """
    if json_api_settings.FORMAT_FIELD_NAMES:
        return format_value(field_name, "underscore")

    return field_name


def format_link_segment(value, format_type=None):
    """
    Takes a string value and returns it with formatted keys as set in `format_type`
    or `JSON_API_FORMAT_RELATED_LINKS`.

    :format_type: Either 'dasherize', 'camelize', 'capitalize' or 'underscore'
    """
    if format_type is None:
        format_type = json_api_settings.FORMAT_RELATED_LINKS
    else:
        warnings.warn(
            DeprecationWarning(
                "Using `format_type` argument is deprecated."
                "Use `format_value` instead."
            )
        )

    return format_value(value, format_type)


def undo_format_link_segment(value):
    """
    Takes a link segment and undos format link segment to underscore which is the Python convention
    but only in case `JSON_API_FORMAT_RELATED_LINKS` is actually configured.
    """

    if json_api_settings.FORMAT_RELATED_LINKS:
        return format_value(value, "underscore")

    return value


def format_value(value, format_type=None):
    if format_type is None:
        warnings.warn(
            DeprecationWarning(
                "Using `format_value` without passing on `format_type` argument is deprecated."
                "Use `format_field_name` instead."
            )
        )
        format_type = json_api_settings.FORMAT_FIELD_NAMES
    if format_type == "dasherize":
        # inflection can't dasherize camelCase
        value = inflection.underscore(value)
        value = inflection.dasherize(value)
    elif format_type == "camelize":
        value = inflection.camelize(value, False)
    elif format_type == "capitalize":
        value = inflection.camelize(value)
    elif format_type == "underscore":
        value = inflection.underscore(value)
    return value


def format_resource_type(value, format_type=None, pluralize=None):
    if format_type is None:
        format_type = json_api_settings.FORMAT_TYPES

    if pluralize is None:
        pluralize = json_api_settings.PLURALIZE_TYPES

    if format_type:
        value = format_value(value, format_type)

    return inflection.pluralize(value) if pluralize else value


def get_related_resource_type(relation):
    from rest_framework_json_api.serializers import PolymorphicModelSerializer

    try:
        return get_resource_type_from_serializer(relation)
    except AttributeError:
        pass
    relation_model = None
    if hasattr(relation, "_meta"):
        relation_model = relation._meta.model
    elif hasattr(relation, "model"):
        # the model type was explicitly passed as a kwarg to ResourceRelatedField
        relation_model = relation.model
    elif hasattr(relation, "get_queryset") and relation.get_queryset() is not None:
        relation_model = relation.get_queryset().model
    elif (
        getattr(relation, "many", False)
        and hasattr(relation.child, "Meta")
        and hasattr(relation.child.Meta, "model")
    ):
        # For ManyToMany relationships, get the model from the child
        # serializer of the list serializer
        relation_model = relation.child.Meta.model
    else:
        parent_serializer = relation.parent
        parent_model = None
        if isinstance(parent_serializer, PolymorphicModelSerializer):
            parent_model = parent_serializer.get_polymorphic_serializer_for_instance(
                parent_serializer.instance
            ).Meta.model
        elif hasattr(parent_serializer, "Meta"):
            parent_model = getattr(parent_serializer.Meta, "model", None)
        elif hasattr(parent_serializer, "parent") and hasattr(
            parent_serializer.parent, "Meta"
        ):
            parent_model = getattr(parent_serializer.parent.Meta, "model", None)

        if parent_model is not None:
            if relation.source:
                if relation.source != "*":
                    parent_model_relation = getattr(parent_model, relation.source)
                else:
                    parent_model_relation = getattr(parent_model, relation.field_name)
            else:
                parent_model_relation = getattr(
                    parent_model, parent_serializer.field_name
                )

            parent_model_relation_type = type(parent_model_relation)
            if parent_model_relation_type is ReverseManyToOneDescriptor:
                relation_model = parent_model_relation.rel.related_model
            elif parent_model_relation_type is ManyToManyDescriptor:
                relation_model = parent_model_relation.field.remote_field.model
                # In case we are in a reverse relation
                if relation_model == parent_model:
                    relation_model = parent_model_relation.field.model
            elif parent_model_relation_type is ReverseGenericManyToOneDescriptor:
                relation_model = parent_model_relation.rel.model
            elif hasattr(parent_model_relation, "field"):
                try:
                    relation_model = parent_model_relation.field.remote_field.model
                except AttributeError:
                    relation_model = parent_model_relation.field.related.model
            else:
                return get_related_resource_type(parent_model_relation)

    if relation_model is None:
        # For ManyRelatedFields on plain Serializers the resource_type
        # cannot be determined from a model, so we must get it from the
        # child_relation
        if hasattr(relation, "child_relation"):
            return get_related_resource_type(relation.child_relation)
        raise APIException(
            _("Could not resolve resource type for relation %s" % relation)
        )

    return get_resource_type_from_model(relation_model)


def get_resource_type_from_model(model):
    json_api_meta = getattr(model, "JSONAPIMeta", None)
    return getattr(json_api_meta, "resource_name", format_resource_type(model.__name__))


def get_resource_type_from_queryset(qs):
    return get_resource_type_from_model(qs.model)


def get_resource_type_from_instance(instance):
    if hasattr(instance, "_meta"):
        return get_resource_type_from_model(instance._meta.model)


def get_resource_type_from_manager(manager):
    return get_resource_type_from_model(manager.model)


def get_resource_type_from_serializer(serializer):
    json_api_meta = getattr(serializer, "JSONAPIMeta", None)
    meta = getattr(serializer, "Meta", None)
    if hasattr(json_api_meta, "resource_name"):
        return json_api_meta.resource_name
    elif hasattr(meta, "resource_name"):
        return meta.resource_name
    elif hasattr(meta, "model"):
        return get_resource_type_from_model(meta.model)
    raise AttributeError(
        f"can not detect 'resource_name' on serializer '{serializer.__class__.__name__}'"
        f" in module '{serializer.__class__.__module__}'"
    )


def get_included_resources(request, serializer=None):
    """Build a list of included resources."""
    include_resources_param = request.query_params.get("include") if request else None
    if include_resources_param:
        return include_resources_param.split(",")
    else:
        return get_default_included_resources_from_serializer(serializer)


def get_default_included_resources_from_serializer(serializer):
    meta = getattr(serializer, "JSONAPIMeta", None)
    if meta is None and getattr(serializer, "many", False):
        meta = getattr(serializer.child, "JSONAPIMeta", None)
    return list(getattr(meta, "included_resources", []))


def get_included_serializers(serializer):
    warnings.warn(
        DeprecationWarning(
            "Using of `get_included_serializers(serializer)` function is deprecated."
            "Use `serializer.included_serializers` instead."
        )
    )

    return getattr(serializer, "included_serializers", dict())


def get_relation_instance(resource_instance, source, serializer):
    try:
        relation_instance = operator.attrgetter(source)(resource_instance)
    except AttributeError:
        # if the field is not defined on the model then we check the serializer
        # and if no value is there we skip over the field completely
        serializer_method = getattr(serializer, source, None)
        if serializer_method and hasattr(serializer_method, "__call__"):
            relation_instance = serializer_method(resource_instance)
        else:
            return False, None

    if isinstance(relation_instance, Manager):
        relation_instance = relation_instance.all()

    return True, relation_instance


def is_relationship_field(field):
    return isinstance(field, (relations.RelatedField, relations.ManyRelatedField))


class Hyperlink(str):
    """
    A string like object that additionally has an associated name.
    We use this for hyperlinked URLs that may render as a named link
    in some contexts, or render as a plain URL in others.

    Comes from Django REST framework 3.2
    https://github.com/tomchristie/django-rest-framework
    """

    def __new__(self, url, name):
        ret = str.__new__(self, url)
        ret.name = name
        return ret

    is_hyperlink = True


def format_drf_errors(response, context, exc):
    errors = []
    # handle generic errors. ValidationError('test') in a view for example
    if isinstance(response.data, list):
        for message in response.data:
            errors.extend(format_error_object(message, "/data", response))
    # handle all errors thrown from serializers
    else:
        # Avoid circular deps
        from rest_framework import generics

        has_serializer = isinstance(context["view"], generics.GenericAPIView)
        if has_serializer:
            serializer = context["view"].get_serializer()
            fields = get_serializer_fields(serializer) or dict()
            relationship_fields = [
                name for name, field in fields.items() if is_relationship_field(field)
            ]

        for field, error in response.data.items():
            field = format_field_name(field)
            pointer = None
            # pointer can be determined only if there's a serializer.
            if has_serializer:
                rel = "relationships" if field in relationship_fields else "attributes"
                pointer = "/data/{}/{}".format(rel, field)
            if isinstance(exc, Http404) and isinstance(error, str):
                # 404 errors don't have a pointer
                errors.extend(format_error_object(error, None, response))
            elif isinstance(error, str):
                classes = inspect.getmembers(exceptions, inspect.isclass)
                # DRF sets the `field` to 'detail' for its own exceptions
                if isinstance(exc, tuple(x[1] for x in classes)):
                    pointer = "/data"
                errors.extend(format_error_object(error, pointer, response))
            elif isinstance(error, list):
                errors.extend(format_error_object(error, pointer, response))
            else:
                errors.extend(format_error_object(error, pointer, response))

    context["view"].resource_name = "errors"
    response.data = errors

    return response


def format_error_object(message, pointer, response):
    errors = []
    if isinstance(message, dict):

        # as there is no required field in error object we check that all fields are string
        # except links and source which might be a dict
        is_custom_error = all(
            [
                isinstance(value, str)
                for key, value in message.items()
                if key not in ["links", "source"]
            ]
        )

        if is_custom_error:
            if "source" not in message:
                message["source"] = {}
            message["source"] = {
                "pointer": pointer,
            }
            errors.append(message)
        else:
            for k, v in message.items():
                errors.extend(
                    format_error_object(v, pointer + "/{}".format(k), response)
                )
    elif isinstance(message, list):
        for num, error in enumerate(message):
            if isinstance(error, (list, dict)):
                new_pointer = pointer + "/{}".format(num)
            else:
                new_pointer = pointer
            if error:
                errors.extend(format_error_object(error, new_pointer, response))
    else:
        error_obj = {
            "detail": message,
            "status": encoding.force_str(response.status_code),
        }
        if pointer is not None:
            error_obj["source"] = {
                "pointer": pointer,
            }
        code = getattr(message, "code", None)
        if code is not None:
            error_obj["code"] = code
        errors.append(error_obj)

    return errors


def format_errors(data):
    if len(data) > 1 and isinstance(data, list):
        data.sort(key=lambda x: x.get("source", {}).get("pointer", ""))
    return {"errors": data}



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=50)),
                ("email", models.EmailField(max_length=254)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="AuthorBio",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("body", models.TextField()),
                (
                    "author",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="bio",
                        to="example.Author",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Blog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                ("tagline", models.TextField()),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Comment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("body", models.TextField()),
                (
                    "author",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="example.Author",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Entry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("headline", models.CharField(max_length=255)),
                ("body_text", models.TextField(null=True)),
                ("pub_date", models.DateField(null=True)),
                ("mod_date", models.DateField(null=True)),
                ("n_comments", models.IntegerField(default=0)),
                ("n_pingbacks", models.IntegerField(default=0)),
                ("rating", models.IntegerField(default=0)),
                ("authors", models.ManyToManyField(to="example.Author")),
                (
                    "blog",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="example.Blog"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="comment",
            name="entry",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="example.Entry"
            ),
        ),
    ]



from django.urls import re_path
from rest_framework import routers

from .api.resources.identity import GenericIdentity, Identity
from example.views import (
    AuthorRelationshipView,
    AuthorViewSet,
    BlogRelationshipView,
    BlogViewSet,
    CommentRelationshipView,
    CommentViewSet,
    CompanyViewset,
    DRFBlogViewSet,
    DRFEntryViewSet,
    EntryRelationshipView,
    EntryViewSet,
    FiltersetEntryViewSet,
    LabResultViewSet,
    NoFiltersetEntryViewSet,
    NonPaginatedEntryViewSet,
    ProjectTypeViewset,
    ProjectViewset,
)

router = routers.DefaultRouter(trailing_slash=False)

router.register(r"blogs", BlogViewSet)
# router to test default DRF blog functionalities
router.register(r"drf-blogs", DRFBlogViewSet, "drf-entry-blog")
router.register(r"entries", EntryViewSet)
# these "flavors" of entries are used for various tests:
router.register(r"nopage-entries", NonPaginatedEntryViewSet, "nopage-entry")
router.register(r"filterset-entries", FiltersetEntryViewSet, "filterset-entry")
router.register(r"nofilterset-entries", NoFiltersetEntryViewSet, "nofilterset-entry")
router.register(r"authors", AuthorViewSet)
router.register(r"comments", CommentViewSet)
router.register(r"companies", CompanyViewset)
router.register(r"projects", ProjectViewset)
router.register(r"project-types", ProjectTypeViewset)
router.register(r"lab-results", LabResultViewSet)

# for the old tests
router.register(r"identities", Identity)

urlpatterns = [
    # old tests
    re_path(
        r"identities/default/(?P<pk>\d+)$",
        GenericIdentity.as_view(),
        name="user-default",
    ),
    re_path(
        r"^entries/(?P<entry_pk>[^/.]+)/blog$",
        BlogViewSet.as_view({"get": "retrieve"}),
        name="entry-blog",
    ),
    re_path(
        r"^entries/(?P<entry_pk>[^/.]+)/comments$",
        CommentViewSet.as_view({"get": "list"}),
        name="entry-comments",
    ),
    re_path(
        r"^entries/(?P<entry_pk>[^/.]+)/suggested/$",
        EntryViewSet.as_view({"get": "list"}),
        name="entry-suggested",
    ),
    re_path(
        r"^drf-entries/(?P<entry_pk>[^/.]+)/suggested/$",
        DRFEntryViewSet.as_view({"get": "list"}),
        name="drf-entry-suggested",
    ),
    re_path(
        r"entries/(?P<entry_pk>[^/.]+)/authors$",
        AuthorViewSet.as_view({"get": "list"}),
        name="entry-authors",
    ),
    re_path(
        r"entries/(?P<entry_pk>[^/.]+)/featured$",
        EntryViewSet.as_view({"get": "retrieve"}),
        name="entry-featured",
    ),
    re_path(
        r"^authors/(?P<pk>[^/.]+)/(?P<related_field>[-\w]+)/$",
        AuthorViewSet.as_view({"get": "retrieve_related"}),
        name="author-related",
    ),
    re_path(
        r"^entries/(?P<pk>[^/.]+)/relationships/(?P<related_field>[\-\w]+)$",
        EntryRelationshipView.as_view(),
        name="entry-relationships",
    ),
    re_path(
        r"^blogs/(?P<pk>[^/.]+)/relationships/(?P<related_field>[^/.]+)$",
        BlogRelationshipView.as_view(),
        name="blog-relationships",
    ),
    re_path(
        r"^comments/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$",
        CommentRelationshipView.as_view(),
        name="comment-relationships",
    ),
    re_path(
        r"^authors/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$",
        AuthorRelationshipView.as_view(),
        name="author-relationships",
    ),
]

urlpatterns += router.urls



# Generated by Django 3.0.9 on 2021-07-14 08:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("example", "0009_labresults_author"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="labresults",
            options={"ordering": ("id",)},
        ),
    ]



import pytest
from django.db import models
from rest_framework import status
from rest_framework.fields import Field
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.views import APIView

from rest_framework_json_api import serializers
from rest_framework_json_api.utils import (
    format_field_name,
    format_field_names,
    format_link_segment,
    format_resource_type,
    format_value,
    get_included_serializers,
    get_related_resource_type,
    get_resource_name,
    get_resource_type_from_serializer,
    undo_format_field_name,
    undo_format_field_names,
    undo_format_link_segment,
)
from tests.models import (
    BasicModel,
    DJAModel,
    ForeignKeySource,
    ForeignKeyTarget,
    ManyToManySource,
    ManyToManyTarget,
)
from tests.serializers import BasicModelSerializer, ManyToManyTargetSerializer


def test_get_resource_name_no_view():
    assert get_resource_name({}) is None


@pytest.mark.parametrize(
    "format_type,pluralize_type,output",
    [
        (False, False, "APIView"),
        (False, True, "APIViews"),
        ("dasherize", False, "api-view"),
        ("dasherize", True, "api-views"),
    ],
)
def test_get_resource_name_from_view(settings, format_type, pluralize_type, output):
    settings.JSON_API_FORMAT_TYPES = format_type
    settings.JSON_API_PLURALIZE_TYPES = pluralize_type

    view = APIView()
    context = {"view": view}
    assert output == get_resource_name(context)


@pytest.mark.parametrize(
    "format_type,pluralize_type",
    [
        (False, False),
        (False, True),
        ("dasherize", False),
        ("dasherize", True),
    ],
)
def test_get_resource_name_from_view_custom_resource_name(
    settings, format_type, pluralize_type
):
    settings.JSON_API_FORMAT_TYPES = format_type
    settings.JSON_API_PLURALIZE_TYPES = pluralize_type

    view = APIView()
    view.resource_name = "custom"
    context = {"view": view}
    assert "custom" == get_resource_name(context)


@pytest.mark.parametrize(
    "format_type,pluralize_type,output",
    [
        (False, False, "BasicModel"),
        (False, True, "BasicModels"),
        ("dasherize", False, "basic-model"),
        ("dasherize", True, "basic-models"),
    ],
)
def test_get_resource_name_from_model(settings, format_type, pluralize_type, output):
    settings.JSON_API_FORMAT_TYPES = format_type
    settings.JSON_API_PLURALIZE_TYPES = pluralize_type

    view = APIView()
    view.model = BasicModel
    context = {"view": view}
    assert output == get_resource_name(context)


@pytest.mark.parametrize(
    "format_type,pluralize_type,output",
    [
        (False, False, "BasicModel"),
        (False, True, "BasicModels"),
        ("dasherize", False, "basic-model"),
        ("dasherize", True, "basic-models"),
    ],
)
def test_get_resource_name_from_model_serializer_class(
    settings, format_type, pluralize_type, output
):
    settings.JSON_API_FORMAT_TYPES = format_type
    settings.JSON_API_PLURALIZE_TYPES = pluralize_type

    view = GenericAPIView()
    view.serializer_class = BasicModelSerializer
    context = {"view": view}
    assert output == get_resource_name(context)


@pytest.mark.parametrize(
    "format_type,pluralize_type",
    [
        (False, False),
        (False, True),
        ("dasherize", False),
        ("dasherize", True),
    ],
)
def test_get_resource_name_from_model_serializer_class_custom_resource_name(
    settings, format_type, pluralize_type
):
    settings.JSON_API_FORMAT_TYPES = format_type
    settings.JSON_API_PLURALIZE_TYPES = pluralize_type

    view = GenericAPIView()
    view.serializer_class = BasicModelSerializer
    view.serializer_class.Meta.resource_name = "custom"

    context = {"view": view}
    assert "custom" == get_resource_name(context)


@pytest.mark.parametrize(
    "format_type,pluralize_type",
    [
        (False, False),
        (False, True),
        ("dasherize", False),
        ("dasherize", True),
    ],
)
def test_get_resource_name_from_plain_serializer_class(
    settings, format_type, pluralize_type
):
    class PlainSerializer(serializers.Serializer):
        class Meta:
            resource_name = "custom"

    settings.JSON_API_FORMAT_TYPES = format_type
    settings.JSON_API_PLURALIZE_TYPES = pluralize_type

    view = GenericAPIView()
    view.serializer_class = PlainSerializer

    context = {"view": view}
    assert "custom" == get_resource_name(context)


@pytest.mark.parametrize(
    "status_code",
    [
        status.HTTP_400_BAD_REQUEST,
        status.HTTP_403_FORBIDDEN,
        status.HTTP_500_INTERNAL_SERVER_ERROR,
    ],
)
def test_get_resource_name_with_errors(status_code):
    view = APIView()
    context = {"view": view}
    view.response = Response(status=status_code)
    assert "errors" == get_resource_name(context)


@pytest.mark.parametrize(
    "format_type,output",
    [
        (False, {"full_name": {"last-name": "a", "first-name": "b"}}),
        ("camelize", {"fullName": {"last-name": "a", "first-name": "b"}}),
        ("capitalize", {"FullName": {"last-name": "a", "first-name": "b"}}),
        ("dasherize", {"full-name": {"last-name": "a", "first-name": "b"}}),
        ("underscore", {"full_name": {"last-name": "a", "first-name": "b"}}),
    ],
)
def test_format_field_names(settings, format_type, output):
    settings.JSON_API_FORMAT_FIELD_NAMES = format_type

    value = {"full_name": {"last-name": "a", "first-name": "b"}}
    assert format_field_names(value, format_type) == output


@pytest.mark.parametrize(
    "format_type,output",
    [
        (False, {"fullName": "Test Name"}),
        ("camelize", {"full_name": "Test Name"}),
    ],
)
def test_undo_format_field_names(settings, format_type, output):
    settings.JSON_API_FORMAT_FIELD_NAMES = format_type

    value = {"fullName": "Test Name"}
    assert undo_format_field_names(value) == output


@pytest.mark.parametrize(
    "format_type,output",
    [
        (False, "full_name"),
        ("camelize", "fullName"),
        ("capitalize", "FullName"),
        ("dasherize", "full-name"),
        ("underscore", "full_name"),
    ],
)
def test_format_field_name(settings, format_type, output):
    settings.JSON_API_FORMAT_FIELD_NAMES = format_type

    field_name = "full_name"
    assert format_field_name(field_name) == output


@pytest.mark.parametrize(
    "format_type,output",
    [
        (False, "fullName"),
        ("camelize", "full_name"),
    ],
)
def test_undo_format_field_name(settings, format_type, output):
    settings.JSON_API_FORMAT_FIELD_NAMES = format_type

    field_name = "fullName"
    assert undo_format_field_name(field_name) == output


@pytest.mark.parametrize(
    "format_type,output",
    [
        (False, "first_Name"),
        ("camelize", "firstName"),
        ("capitalize", "FirstName"),
        ("dasherize", "first-name"),
        ("underscore", "first_name"),
    ],
)
def test_format_link_segment(settings, format_type, output):
    settings.JSON_API_FORMAT_RELATED_LINKS = format_type
    assert format_link_segment("first_Name") == output


def test_format_link_segment_deprecates_format_type_argument():
    with pytest.deprecated_call():
        assert "first-name" == format_link_segment("first_name", "dasherize")


@pytest.mark.parametrize(
    "format_links,output",
    [
        (False, "fullName"),
        ("camelize", "full_name"),
    ],
)
def test_undo_format_link_segment(settings, format_links, output):
    settings.JSON_API_FORMAT_RELATED_LINKS = format_links

    link_segment = "fullName"
    assert undo_format_link_segment(link_segment) == output


@pytest.mark.parametrize(
    "format_type,output",
    [
        (False, "first_name"),
        ("camelize", "firstName"),
        ("capitalize", "FirstName"),
        ("dasherize", "first-name"),
        ("underscore", "first_name"),
    ],
)
def test_format_value(settings, format_type, output):
    assert format_value("first_name", format_type) == output


def test_format_value_deprecates_default_format_type_argument():
    with pytest.deprecated_call():
        assert "first_name" == format_value("first_name")


@pytest.mark.parametrize(
    "resource_type,pluralize,output",
    [
        (None, None, "ResourceType"),
        ("camelize", False, "resourceType"),
        ("camelize", True, "resourceTypes"),
    ],
)
def test_format_resource_type(settings, resource_type, pluralize, output):
    assert format_resource_type("ResourceType", resource_type, pluralize) == output


@pytest.mark.parametrize(
    "model_class,field,output",
    [
        (ManyToManySource, "targets", "ManyToManyTarget"),
        (ManyToManyTarget, "sources", "ManyToManySource"),
        (ForeignKeySource, "target", "ForeignKeyTarget"),
        (ForeignKeyTarget, "sources", "ForeignKeySource"),
    ],
)
def test_get_related_resource_type(model_class, field, output):
    class RelatedResourceTypeSerializer(serializers.ModelSerializer):
        class Meta:
            model = model_class
            fields = (field,)

    serializer = RelatedResourceTypeSerializer()
    field = serializer.fields[field]
    assert get_related_resource_type(field) == output


@pytest.mark.parametrize(
    "related_field_kwargs,output",
    [
        ({"queryset": BasicModel.objects}, "BasicModel"),
        ({"queryset": BasicModel.objects, "model": BasicModel}, "BasicModel"),
        ({"model": BasicModel, "read_only": True}, "BasicModel"),
    ],
)
def test_get_related_resource_type_from_plain_serializer_class(
    related_field_kwargs, output
):
    class PlainRelatedResourceTypeSerializer(serializers.Serializer):
        basic_models = serializers.ResourceRelatedField(
            many=True, **related_field_kwargs
        )

    serializer = PlainRelatedResourceTypeSerializer()
    field = serializer.fields["basic_models"]
    assert get_related_resource_type(field) == output


def test_get_included_serializers():
    class DeprecatedIncludedSerializersModel(DJAModel):
        self = models.ForeignKey("self", on_delete=models.CASCADE)
        target = models.ForeignKey(ManyToManyTarget, on_delete=models.CASCADE)
        other_target = models.ForeignKey(ManyToManyTarget, on_delete=models.CASCADE)

        class Meta:
            app_label = "tests"

    class DeprecatedIncludedSerializersSerializer(serializers.ModelSerializer):
        included_serializers = {
            "self": "self",
            "target": ManyToManyTargetSerializer,
            "other_target": "tests.serializers.ManyToManyTargetSerializer",
        }

        class Meta:
            model = DeprecatedIncludedSerializersModel
            fields = ("self", "other_target", "target")

    with pytest.deprecated_call():
        included_serializers = get_included_serializers(
            DeprecatedIncludedSerializersSerializer
        )

    expected_included_serializers = {
        "self": DeprecatedIncludedSerializersSerializer,
        "target": ManyToManyTargetSerializer,
        "other_target": ManyToManyTargetSerializer,
    }

    assert included_serializers == expected_included_serializers


def test_get_resource_type_from_serializer_without_resource_name_raises_error():
    class SerializerWithoutResourceName(serializers.Serializer):
        something = Field()

    serializer = SerializerWithoutResourceName()

    with pytest.raises(AttributeError) as excinfo:
        get_resource_type_from_serializer(serializer=serializer)
    assert str(excinfo.value) == (
        "can not detect 'resource_name' on serializer "
        "'SerializerWithoutResourceName' in module 'tests.test_utils'"
    )



import pytest
from django.contrib.auth import get_user_model

from rest_framework_json_api import serializers
from rest_framework_json_api.renderers import JSONRenderer
from rest_framework_json_api.utils import get_serializer_fields

pytestmark = pytest.mark.django_db


class ResourceSerializer(serializers.ModelSerializer):
    version = serializers.SerializerMethodField()

    def get_version(self, obj):
        return "1.0.0"

    class Meta:
        fields = ("username",)
        meta_fields = ("version",)
        model = get_user_model()


def test_build_json_resource_obj():
    resource = {"username": "Alice", "version": "1.0.0"}

    serializer = ResourceSerializer(data={"username": "Alice"})
    serializer.is_valid()
    resource_instance = serializer.save()

    output = {
        "type": "user",
        "id": "1",
        "attributes": {"username": "Alice"},
        "meta": {"version": "1.0.0"},
    }

    assert (
        JSONRenderer.build_json_resource_obj(
            get_serializer_fields(serializer),
            resource,
            resource_instance,
            "user",
            serializer,
        )
        == output
    )


def test_can_override_methods():
    """
    Make sure extract_attributes and extract_relationships can be overriden.
    """

    resource = {"username": "Alice", "version": "1.0.0"}
    serializer = ResourceSerializer(data={"username": "Alice"})
    serializer.is_valid()
    resource_instance = serializer.save()

    output = {
        "type": "user",
        "id": "1",
        "attributes": {"username": "Alice"},
        "meta": {"version": "1.0.0"},
    }

    class CustomRenderer(JSONRenderer):
        extract_attributes_was_overriden = False
        extract_relationships_was_overriden = False

        @classmethod
        def extract_attributes(cls, fields, resource):
            cls.extract_attributes_was_overriden = True
            return super().extract_attributes(fields, resource)

        @classmethod
        def extract_relationships(cls, fields, resource, resource_instance):
            cls.extract_relationships_was_overriden = True
            return super().extract_relationships(fields, resource, resource_instance)

    assert (
        CustomRenderer.build_json_resource_obj(
            get_serializer_fields(serializer),
            resource,
            resource_instance,
            "user",
            serializer,
        )
        == output
    )
    assert CustomRenderer.extract_attributes_was_overriden
    assert CustomRenderer.extract_relationships_was_overriden


def test_extract_attributes():
    fields = {
        "id": serializers.Field(),
        "username": serializers.Field(),
        "deleted": serializers.ReadOnlyField(),
    }
    resource = {"id": 1, "deleted": None, "username": "jerel"}
    expected = {"username": "jerel", "deleted": None}
    assert sorted(JSONRenderer.extract_attributes(fields, resource)) == sorted(
        expected
    ), "Regular fields should be extracted"
    assert sorted(JSONRenderer.extract_attributes(fields, {})) == sorted(
        {"username": ""}
    ), "Should not extract read_only fields on empty serializer"


def test_extract_meta():
    serializer = ResourceSerializer(data={"username": "jerel", "version": "1.0.0"})
    serializer.is_valid()
    expected = {
        "version": "1.0.0",
    }
    assert JSONRenderer.extract_meta(serializer, serializer.data) == expected


class ExtractRootMetaResourceSerializer(ResourceSerializer):
    def get_root_meta(self, resource, many):
        if many:
            return {"foo": "meta-many-value"}
        else:
            return {"foo": "meta-value"}


class InvalidExtractRootMetaResourceSerializer(ResourceSerializer):
    def get_root_meta(self, resource, many):
        return "not a dict"


def test_extract_root_meta():
    serializer = ExtractRootMetaResourceSerializer()
    expected = {
        "foo": "meta-value",
    }
    assert JSONRenderer.extract_root_meta(serializer, {}) == expected


def test_extract_root_meta_many():
    serializer = ExtractRootMetaResourceSerializer(many=True)
    expected = {"foo": "meta-many-value"}
    assert JSONRenderer.extract_root_meta(serializer, {}) == expected


def test_extract_root_meta_invalid_meta():
    serializer = InvalidExtractRootMetaResourceSerializer()
    with pytest.raises(AssertionError):
        JSONRenderer.extract_root_meta(serializer, {})



import pytest
from django.db import models

from rest_framework_json_api import serializers
from tests.models import DJAModel, ManyToManyTarget
from tests.serializers import ManyToManyTargetSerializer


def test_get_included_serializers():
    class IncludedSerializersModel(DJAModel):
        self = models.ForeignKey("self", on_delete=models.CASCADE)
        target = models.ForeignKey(ManyToManyTarget, on_delete=models.CASCADE)
        other_target = models.ForeignKey(ManyToManyTarget, on_delete=models.CASCADE)

        class Meta:
            app_label = "tests"

    class IncludedSerializersSerializer(serializers.ModelSerializer):
        included_serializers = {
            "self": "self",
            "target": ManyToManyTargetSerializer,
            "other_target": "tests.serializers.ManyToManyTargetSerializer",
        }

        class Meta:
            model = IncludedSerializersModel
            fields = ("self", "other_target", "target")

    included_serializers = IncludedSerializersSerializer.included_serializers
    expected_included_serializers = {
        "self": IncludedSerializersSerializer,
        "target": ManyToManyTargetSerializer,
        "other_target": ManyToManyTargetSerializer,
    }

    assert included_serializers == expected_included_serializers


def test_reserved_field_names():
    with pytest.raises(AssertionError) as e:

        class ReservedFieldNamesSerializer(serializers.Serializer):
            meta = serializers.CharField()
            results = serializers.CharField()

        ReservedFieldNamesSerializer().fields

    assert str(e.value) == (
        "Serializer class tests.test_serializers.test_reserved_field_names.<locals>."
        "ReservedFieldNamesSerializer uses following reserved field name(s) which is "
        "not allowed: meta, results"
    )


def test_serializer_fields_deprecated_field_name_type():
    with pytest.deprecated_call():

        class TypeFieldNameSerializer(serializers.Serializer):
            type = serializers.CharField()

        TypeFieldNameSerializer().fields



from django.contrib.auth import models as auth_models
from rest_framework import serializers


class IdentitySerializer(serializers.ModelSerializer):
    """
    Identity Serializer
    """

    def validate_first_name(self, data):
        if len(data) > 10:
            raise serializers.ValidationError("There's a problem with first name")
        return data

    def validate_last_name(self, data):
        if len(data) > 10:
            raise serializers.ValidationError(
                {
                    "id": "armageddon101",
                    "detail": "Hey! You need a last name!",
                    "meta": "something",
                }
            )
        return data

    class Meta:
        model = auth_models.User
        fields = (
            "id",
            "first_name",
            "last_name",
            "email",
        )



from django.test import override_settings
from django.urls import reverse

from example.tests import TestBase


class GenericViewSet(TestBase):
    """
    Test expected responses coming from a Generic ViewSet
    """

    def test_default_rest_framework_behavior(self):
        """
        This is more of an example really, showing default behavior
        """
        url = reverse("user-default", kwargs={"pk": self.miles.pk})

        response = self.client.get(url)

        self.assertEqual(200, response.status_code)

        expected = {
            "id": 2,
            "first_name": "Miles",
            "last_name": "Davis",
            "email": "miles@example.com",
        }

        assert expected == response.json()

    def test_ember_expected_renderer(self):
        """
        The :class:`UserEmber` ViewSet has the ``resource_name`` of 'data'
        so that should be the key in the JSON response.
        """
        url = reverse("user-manual-resource-name", kwargs={"pk": self.miles.pk})

        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.get(url)
        self.assertEqual(200, response.status_code)

        expected = {
            "data": {
                "type": "data",
                "id": "2",
                "attributes": {
                    "first-name": "Miles",
                    "last-name": "Davis",
                    "email": "miles@example.com",
                },
            }
        }

        assert expected == response.json()

    def test_default_validation_exceptions(self):
        """
        Default validation exceptions should conform to JSON:API spec
        """
        expected = {
            "errors": [
                {
                    "status": "400",
                    "source": {
                        "pointer": "/data/attributes/email",
                    },
                    "detail": "Enter a valid email address.",
                    "code": "invalid",
                },
                {
                    "status": "400",
                    "source": {
                        "pointer": "/data/attributes/first-name",
                    },
                    "detail": "There's a problem with first name",
                    "code": "invalid",
                },
            ]
        }
        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.post(
                "/identities",
                {
                    "data": {
                        "type": "users",
                        "attributes": {
                            "email": "bar",
                            "first_name": "alajflajaljalajlfjafljalj",
                        },
                    }
                },
            )

        assert expected == response.json()

    def test_custom_validation_exceptions(self):
        """
        Exceptions should be able to be formatted manually
        """
        expected = {
            "errors": [
                {
                    "status": "400",
                    "source": {
                        "pointer": "/data/attributes/email",
                    },
                    "detail": "Enter a valid email address.",
                    "code": "invalid",
                },
                {
                    "id": "armageddon101",
                    "detail": "Hey! You need a last name!",
                    "meta": "something",
                    "source": {"pointer": "/data/attributes/lastName"},
                },
            ]
        }
        response = self.client.post(
            "/identities",
            {
                "data": {
                    "type": "users",
                    "attributes": {
                        "email": "bar",
                        "last_name": "alajflajaljalajlfjafljalj",
                    },
                }
            },
        )

        assert expected == response.json()



import json
from datetime import datetime

import pytest
from django.urls import reverse
from rest_framework import viewsets
from rest_framework.serializers import ModelSerializer, SerializerMethodField

from rest_framework_json_api.renderers import JSONRenderer

from example.models import Blog, Comment, Entry


# serializers
class RelatedModelSerializer(ModelSerializer):
    class Meta:
        model = Comment
        fields = ("id",)


class DummyTestSerializer(ModelSerializer):
    """
    This serializer is a simple compound document serializer which includes only
    a single embedded relation
    """

    related_models = RelatedModelSerializer(
        source="comments", many=True, read_only=True
    )

    json_field = SerializerMethodField()

    def get_json_field(self, entry):
        return {"JsonKey": "JsonValue"}

    class Meta:
        model = Entry
        fields = ("related_models", "json_field")


# views
class DummyTestViewSet(viewsets.ModelViewSet):
    queryset = Entry.objects.all()
    serializer_class = DummyTestSerializer


def render_dummy_test_serialized_view(view_class):
    serializer = DummyTestSerializer(instance=Entry())
    renderer = JSONRenderer()
    return renderer.render(serializer.data, renderer_context={"view": view_class()})


# tests
def test_simple_reverse_relation_included_renderer():
    """
    Test renderer when a single reverse fk relation is passed.
    """
    rendered = render_dummy_test_serialized_view(DummyTestViewSet)

    assert rendered


def test_render_format_field_names(settings):
    """Test that json field is kept untouched."""
    settings.JSON_API_FORMAT_FIELD_NAMES = "dasherize"
    rendered = render_dummy_test_serialized_view(DummyTestViewSet)

    result = json.loads(rendered.decode())
    assert result["data"]["attributes"]["json-field"] == {"JsonKey": "JsonValue"}


@pytest.mark.django_db
def test_blog_create(client):

    url = reverse("drf-entry-blog-list")
    name = "Dummy Name"

    request_data = {
        "data": {"attributes": {"name": name}, "type": "blogs"},
    }

    resp = client.post(url, request_data)

    # look for created blog in database
    blog = Blog.objects.filter(name=name)

    # check if blog exists in database
    assert blog.count() == 1

    # get created blog from database
    blog = blog.first()

    expected = {
        "data": {
            "attributes": {"name": blog.name, "tags": []},
            "id": "{}".format(blog.id),
            "links": {"self": "http://testserver/blogs/{}".format(blog.id)},
            "meta": {"copyright": datetime.now().year},
            "type": "blogs",
        },
        "meta": {"apiDocs": "/docs/api/blogs"},
    }

    assert resp.status_code == 201
    assert resp.json() == expected


@pytest.mark.django_db
def test_get_object_gives_correct_blog(client, blog, entry):

    url = reverse("drf-entry-blog-detail", kwargs={"entry_pk": entry.id})
    resp = client.get(url)
    expected = {
        "data": {
            "attributes": {"name": blog.name, "tags": []},
            "id": "{}".format(blog.id),
            "links": {"self": "http://testserver/blogs/{}".format(blog.id)},
            "meta": {"copyright": datetime.now().year},
            "type": "blogs",
        },
        "meta": {"apiDocs": "/docs/api/blogs"},
    }
    got = resp.json()
    assert got == expected


@pytest.mark.django_db
def test_get_object_patches_correct_blog(client, blog, entry):

    url = reverse("drf-entry-blog-detail", kwargs={"entry_pk": entry.id})
    new_name = blog.name + " update"
    assert not new_name == blog.name

    request_data = {
        "data": {
            "attributes": {"name": new_name},
            "id": "{}".format(blog.id),
            "links": {"self": "http://testserver/blogs/{}".format(blog.id)},
            "meta": {"copyright": datetime.now().year},
            "relationships": {"tags": {"data": []}},
            "type": "blogs",
        },
        "meta": {"apiDocs": "/docs/api/blogs"},
    }

    resp = client.patch(url, data=request_data)

    assert resp.status_code == 200

    expected = {
        "data": {
            "attributes": {"name": new_name, "tags": []},
            "id": "{}".format(blog.id),
            "links": {"self": "http://testserver/blogs/{}".format(blog.id)},
            "meta": {"copyright": datetime.now().year},
            "type": "blogs",
        },
        "meta": {"apiDocs": "/docs/api/blogs"},
    }
    got = resp.json()
    assert got == expected


@pytest.mark.django_db
def test_get_object_deletes_correct_blog(client, entry):

    url = reverse("drf-entry-blog-detail", kwargs={"entry_pk": entry.id})

    resp = client.delete(url)

    assert resp.status_code == 204


@pytest.mark.django_db
def test_get_entry_list_with_blogs(client, entry):
    url = reverse("drf-entry-suggested", kwargs={"entry_pk": entry.id})
    resp = client.get(url)

    got = resp.json()

    expected = {
        "links": {
            "first": "http://testserver/drf-entries/1/suggested/?page%5Bnumber%5D=1",
            "last": "http://testserver/drf-entries/1/suggested/?page%5Bnumber%5D=1",
            "next": None,
            "prev": None,
        },
        "data": [
            {
                "type": "entries",
                "id": "1",
                "attributes": {
                    "tags": [],
                },
                "links": {"self": "http://testserver/drf-blogs/1"},
            }
        ],
        "meta": {"pagination": {"page": 1, "pages": 1, "count": 1}},
    }

    assert resp.status_code == 200
    assert got == expected



import json

import pytest
from django.utils import timezone

from rest_framework_json_api import serializers, views
from rest_framework_json_api.renderers import JSONRenderer

from example.models import Author, Blog, Comment, Entry


# serializers
class RelatedModelSerializer(serializers.ModelSerializer):
    blog = serializers.ReadOnlyField(source="entry.blog")

    class Meta:
        model = Comment
        fields = ("id", "blog")


class DummyTestSerializer(serializers.ModelSerializer):
    """
    This serializer is a simple compound document serializer which includes only
    a single embedded relation
    """

    related_models = RelatedModelSerializer(
        source="comments", many=True, read_only=True
    )

    json_field = serializers.SerializerMethodField()

    def get_json_field(self, entry):
        return {"JsonKey": "JsonValue"}

    class Meta:
        model = Entry
        fields = ("related_models", "json_field")

    class JSONAPIMeta:
        included_resources = ("related_models",)


class EntryDRFSerializers(serializers.ModelSerializer):
    class Meta:
        model = Entry
        fields = ("headline", "body_text")
        read_only_fields = ("tags",)


class CommentWithNestedFieldsSerializer(serializers.ModelSerializer):
    entry = EntryDRFSerializers()

    class Meta:
        model = Comment
        exclude = ("created_at", "modified_at", "author")
        # fields = ('entry', 'body', 'author',)


class AuthorWithNestedFieldsSerializer(serializers.ModelSerializer):
    comments = CommentWithNestedFieldsSerializer(many=True)

    class Meta:
        model = Author
        fields = ("name", "email", "comments")


# views
class DummyTestViewSet(views.ModelViewSet):
    queryset = Entry.objects.all()
    serializer_class = DummyTestSerializer


class ReadOnlyDummyTestViewSet(views.ReadOnlyModelViewSet):
    queryset = Entry.objects.all()
    serializer_class = DummyTestSerializer


class AuthorWithNestedFieldsViewSet(views.ModelViewSet):
    queryset = Author.objects.all()
    serializer_class = AuthorWithNestedFieldsSerializer
    resource_name = "authors"


def render_dummy_test_serialized_view(view_class, instance):
    serializer = view_class.serializer_class(instance=instance)
    renderer = JSONRenderer()
    return renderer.render(serializer.data, renderer_context={"view": view_class()})


def test_simple_reverse_relation_included_renderer():
    """
    Test renderer when a single reverse fk relation is passed.
    """
    rendered = render_dummy_test_serialized_view(DummyTestViewSet, Entry())

    assert rendered


def test_simple_reverse_relation_included_read_only_viewset():
    rendered = render_dummy_test_serialized_view(ReadOnlyDummyTestViewSet, Entry())

    assert rendered


def test_render_format_field_names(settings):
    """Test that json field is kept untouched."""
    settings.JSON_API_FORMAT_FIELD_NAMES = "dasherize"
    rendered = render_dummy_test_serialized_view(DummyTestViewSet, Entry())

    result = json.loads(rendered.decode())
    assert result["data"]["attributes"]["json-field"] == {"JsonKey": "JsonValue"}


def test_writeonly_not_in_response():
    """Test that writeonly fields are not shown in list response"""

    class WriteonlyTestSerializer(serializers.ModelSerializer):
        """Serializer for testing the absence of write_only fields"""

        comments = serializers.ResourceRelatedField(
            many=True, write_only=True, queryset=Comment.objects.all()
        )

        rating = serializers.IntegerField(write_only=True)

        class Meta:
            model = Entry
            fields = ("comments", "rating")

    class WriteOnlyDummyTestViewSet(views.ReadOnlyModelViewSet):
        queryset = Entry.objects.all()
        serializer_class = WriteonlyTestSerializer

    rendered = render_dummy_test_serialized_view(WriteOnlyDummyTestViewSet, Entry())
    result = json.loads(rendered.decode())

    assert "rating" not in result["data"]["attributes"]
    assert "relationships" not in result["data"]


def test_render_empty_relationship_reverse_lookup():
    """Test that empty relationships are rendered as None."""

    class EmptyRelationshipSerializer(serializers.ModelSerializer):
        class Meta:
            model = Author
            fields = ("bio",)

    class EmptyRelationshipViewSet(views.ReadOnlyModelViewSet):
        queryset = Author.objects.all()
        serializer_class = EmptyRelationshipSerializer

    rendered = render_dummy_test_serialized_view(EmptyRelationshipViewSet, Author())
    result = json.loads(rendered.decode())
    assert "relationships" in result["data"]
    assert "bio" in result["data"]["relationships"]
    assert result["data"]["relationships"]["bio"] == {"data": None}


@pytest.mark.django_db
def test_extract_relation_instance(comment):
    serializer = RelatedModelSerializer(instance=comment)

    got = JSONRenderer.extract_relation_instance(
        field=serializer.fields["blog"], resource_instance=comment
    )
    assert got == comment.entry.blog


def test_render_serializer_as_attribute(db):
    # setting up
    blog = Blog.objects.create(name="Some Blog", tagline="It's a blog")
    entry = Entry.objects.create(
        blog=blog,
        headline="headline",
        body_text="body_text",
        pub_date=timezone.now(),
        mod_date=timezone.now(),
        n_comments=0,
        n_pingbacks=0,
        rating=3,
    )

    author = Author.objects.create(name="some_author", email="some_author@example.org")
    entry.authors.add(author)

    Comment.objects.create(
        entry=entry, body="testing one two three", author=Author.objects.first()
    )

    rendered = render_dummy_test_serialized_view(AuthorWithNestedFieldsViewSet, author)
    result = json.loads(rendered.decode())

    expected = {
        "data": {
            "type": "authors",
            "id": "1",
            "attributes": {
                "name": "some_author",
                "email": "some_author@example.org",
                "comments": [
                    {
                        "id": 1,
                        "entry": {
                            "headline": "headline",
                            "body_text": "body_text",
                        },
                        "body": "testing one two three",
                    }
                ],
            },
        }
    }
    assert expected == result



import random

import pytest
from django.urls import reverse
from rest_framework import status

from example.factories import ArtProjectFactory, ProjectTypeFactory

pytestmark = pytest.mark.django_db


def test_polymorphism_on_detail(single_art_project, client):
    response = client.get(
        reverse("project-detail", kwargs={"pk": single_art_project.pk})
    )
    content = response.json()
    assert content["data"]["type"] == "artProjects"


def test_polymorphism_on_detail_relations(single_company, client):
    response = client.get(reverse("company-detail", kwargs={"pk": single_company.pk}))
    content = response.json()
    assert (
        content["data"]["relationships"]["currentProject"]["data"]["type"]
        == "artProjects"
    )
    assert {
        rel["type"]
        for rel in content["data"]["relationships"]["futureProjects"]["data"]
    } == {"researchProjects", "artProjects"}


def test_polymorphism_on_included_relations(single_company, client):
    response = client.get(
        reverse("company-detail", kwargs={"pk": single_company.pk})
        + "?include=current_project,future_projects,current_art_project,current_research_project"
    )
    content = response.json()
    assert (
        content["data"]["relationships"]["currentProject"]["data"]["type"]
        == "artProjects"
    )
    assert (
        content["data"]["relationships"]["currentArtProject"]["data"]["type"]
        == "artProjects"
    )
    assert content["data"]["relationships"]["currentResearchProject"]["data"] is None
    assert {
        rel["type"]
        for rel in content["data"]["relationships"]["futureProjects"]["data"]
    } == {"researchProjects", "artProjects"}
    assert {x.get("type") for x in content.get("included")} == {
        "artProjects",
        "artProjects",
        "researchProjects",
    }, "Detail included types are incorrect"
    # Ensure that the child fields are present.
    assert content.get("included")[0].get("attributes").get("artist") is not None
    assert content.get("included")[1].get("attributes").get("artist") is not None
    assert content.get("included")[2].get("attributes").get("supervisor") is not None


def test_polymorphism_on_polymorphic_model_detail_patch(single_art_project, client):
    url = reverse("project-detail", kwargs={"pk": single_art_project.pk})
    response = client.get(url)
    content = response.json()
    test_topic = "test-{}".format(random.randint(0, 999999))
    test_artist = "test-{}".format(random.randint(0, 999999))
    content["data"]["attributes"]["topic"] = test_topic
    content["data"]["attributes"]["artist"] = test_artist
    response = client.patch(url, data=content)
    new_content = response.json()
    assert new_content["data"]["type"] == "artProjects"
    assert new_content["data"]["attributes"]["topic"] == test_topic
    assert new_content["data"]["attributes"]["artist"] == test_artist


def test_patch_on_polymorphic_model_without_including_required_field(
    single_art_project, client
):
    url = reverse("project-detail", kwargs={"pk": single_art_project.pk})
    data = {
        "data": {
            "id": single_art_project.pk,
            "type": "artProjects",
            "attributes": {"description": "New description"},
        }
    }
    response = client.patch(url, data)
    assert response.status_code == status.HTTP_200_OK
    assert response.json()["data"]["attributes"]["description"] == "New description"


def test_polymorphism_on_polymorphic_model_list_post(client):
    test_topic = "New test topic {}".format(random.randint(0, 999999))
    test_artist = "test-{}".format(random.randint(0, 999999))
    test_project_type = ProjectTypeFactory()
    url = reverse("project-list")
    data = {
        "data": {
            "type": "artProjects",
            "attributes": {"topic": test_topic, "artist": test_artist},
            "relationships": {
                "projectType": {
                    "data": {"type": "projectTypes", "id": test_project_type.pk}
                }
            },
        }
    }
    response = client.post(url, data=data)
    content = response.json()
    assert content["data"]["id"] is not None
    assert content["data"]["type"] == "artProjects"
    assert content["data"]["attributes"]["topic"] == test_topic
    assert content["data"]["attributes"]["artist"] == test_artist
    assert content["data"]["relationships"]["projectType"]["data"]["id"] == str(
        test_project_type.pk
    )


def test_polymorphism_on_polymorphic_model_w_included_serializers(client):
    test_project = ArtProjectFactory()
    query = "?include=projectType"
    url = reverse("project-list")
    response = client.get(url + query)
    content = response.json()
    assert content["data"][0]["id"] == str(test_project.pk)
    assert content["data"][0]["type"] == "artProjects"
    assert content["data"][0]["relationships"]["projectType"]["data"]["id"] == str(
        test_project.project_type.pk
    )
    assert content["included"][0]["type"] == "projectTypes"
    assert content["included"][0]["id"] == str(test_project.project_type.pk)


def test_polymorphic_model_without_any_instance(client):
    expected = {
        "links": {
            "first": "http://testserver/projects?page%5Bnumber%5D=1",
            "last": "http://testserver/projects?page%5Bnumber%5D=1",
            "next": None,
            "prev": None,
        },
        "data": [],
        "meta": {"pagination": {"page": 1, "pages": 1, "count": 0}},
    }

    response = client.get(reverse("project-list"))
    assert response.status_code == 200
    content = response.json()
    assert expected == content


def test_invalid_type_on_polymorphic_model(client):
    test_topic = "New test topic {}".format(random.randint(0, 999999))
    test_artist = "test-{}".format(random.randint(0, 999999))
    url = reverse("project-list")
    data = {
        "data": {
            "type": "invalidProjects",
            "attributes": {"topic": test_topic, "artist": test_artist},
        }
    }
    response = client.post(url, data=data)
    assert response.status_code == 409
    content = response.json()
    assert len(content["errors"]) == 1
    assert content["errors"][0]["status"] == "409"
    try:
        assert (
            content["errors"][0]["detail"]
            == "The resource object's type (invalidProjects) is not the type that constitute the "
            "collection represented by the endpoint (one of [researchProjects, artProjects])."
        )
    except AssertionError:
        # Available type list order isn't enforced
        assert (
            content["errors"][0]["detail"]
            == "The resource object's type (invalidProjects) is not the type that constitute the "
            "collection represented by the endpoint (one of [artProjects, researchProjects])."
        )


def test_polymorphism_relations_update(
    single_company, research_project_factory, client
):
    response = client.get(reverse("company-detail", kwargs={"pk": single_company.pk}))
    content = response.json()
    assert (
        content["data"]["relationships"]["currentProject"]["data"]["type"]
        == "artProjects"
    )

    research_project = research_project_factory()
    content["data"]["relationships"]["currentProject"]["data"] = {
        "type": "researchProjects",
        "id": research_project.pk,
    }
    response = client.patch(
        reverse("company-detail", kwargs={"pk": single_company.pk}), data=content
    )
    assert response.status_code == 200
    content = response.json()
    assert (
        content["data"]["relationships"]["currentProject"]["data"]["type"]
        == "researchProjects"
    )
    assert (
        int(content["data"]["relationships"]["currentProject"]["data"]["id"])
        == research_project.pk
    )


def test_polymorphism_relations_put_405(
    single_company, research_project_factory, client
):
    response = client.get(reverse("company-detail", kwargs={"pk": single_company.pk}))
    content = response.json()
    assert (
        content["data"]["relationships"]["currentProject"]["data"]["type"]
        == "artProjects"
    )

    research_project = research_project_factory()
    content["data"]["relationships"]["currentProject"]["data"] = {
        "type": "researchProjects",
        "id": research_project.pk,
    }
    response = client.put(
        reverse("company-detail", kwargs={"pk": single_company.pk}), data=content
    )
    assert response.status_code == 405


def test_invalid_type_on_polymorphic_relation(
    single_company, research_project_factory, client
):
    response = client.get(reverse("company-detail", kwargs={"pk": single_company.pk}))
    content = response.json()
    assert (
        content["data"]["relationships"]["currentProject"]["data"]["type"]
        == "artProjects"
    )

    research_project = research_project_factory()
    content["data"]["relationships"]["currentProject"]["data"] = {
        "type": "invalidProjects",
        "id": research_project.pk,
    }
    response = client.patch(
        reverse("company-detail", kwargs={"pk": single_company.pk}), data=content
    )
    assert response.status_code == 409
    content = response.json()
    assert len(content["errors"]) == 1
    assert content["errors"][0]["status"] == "409"
    try:
        assert (
            content["errors"][0]["detail"]
            == "Incorrect relation type. Expected one of [researchProjects, artProjects], "
            "received invalidProjects."
        )
    except AssertionError:
        # Available type list order isn't enforced
        assert (
            content["errors"][0]["detail"]
            == "Incorrect relation type. Expected one of [artProjects, researchProjects], "
            "received invalidProjects."
        )



from rest_framework import serializers

from rest_framework_json_api.relations import SerializerMethodResourceRelatedField

from example.models import Blog, Entry


def test_method_name_default():
    class BlogSerializer(serializers.ModelSerializer):
        one_entry = SerializerMethodResourceRelatedField(model=Entry)

        class Meta:
            model = Blog
            fields = ["one_entry"]

        def get_one_entry(self, instance):
            return Entry(id=100)

    serializer = BlogSerializer(instance=Blog())
    assert serializer.data["one_entry"]["id"] == "100"


def test_method_name_custom():
    class BlogSerializer(serializers.ModelSerializer):
        one_entry = SerializerMethodResourceRelatedField(
            model=Entry, method_name="get_custom_entry"
        )

        class Meta:
            model = Blog
            fields = ["one_entry"]

        def get_custom_entry(self, instance):
            return Entry(id=100)

    serializer = BlogSerializer(instance=Blog())
    assert serializer.data["one_entry"]["id"] == "100"



"""
Pagination fields
"""
from collections import OrderedDict

from rest_framework.pagination import LimitOffsetPagination, PageNumberPagination
from rest_framework.utils.urls import remove_query_param, replace_query_param
from rest_framework.views import Response


class JsonApiPageNumberPagination(PageNumberPagination):
    """
    A JSON:API compatible pagination format.
    """

    page_query_param = "page[number]"
    page_size_query_param = "page[size]"
    max_page_size = 100

    def build_link(self, index):
        if not index:
            return None
        url = self.request and self.request.build_absolute_uri() or ""
        return replace_query_param(url, self.page_query_param, index)

    def get_paginated_response(self, data):
        next = None
        previous = None

        if self.page.has_next():
            next = self.page.next_page_number()
        if self.page.has_previous():
            previous = self.page.previous_page_number()

        return Response(
            {
                "results": data,
                "meta": {
                    "pagination": OrderedDict(
                        [
                            ("page", self.page.number),
                            ("pages", self.page.paginator.num_pages),
                            ("count", self.page.paginator.count),
                        ]
                    )
                },
                "links": OrderedDict(
                    [
                        ("first", self.build_link(1)),
                        ("last", self.build_link(self.page.paginator.num_pages)),
                        ("next", self.build_link(next)),
                        ("prev", self.build_link(previous)),
                    ]
                ),
            }
        )


class JsonApiLimitOffsetPagination(LimitOffsetPagination):
    """
    A limit/offset based style. For example:

    .. code::

        http://api.example.org/accounts/?page[limit]=100
        http://api.example.org/accounts/?page[offset]=400&page[limit]=100

    """

    limit_query_param = "page[limit]"
    offset_query_param = "page[offset]"
    max_limit = 100

    def get_last_link(self):
        if self.count == 0:
            return None

        url = self.request.build_absolute_uri()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        offset = (self.count // self.limit) * self.limit

        if offset <= 0:
            return remove_query_param(url, self.offset_query_param)

        return replace_query_param(url, self.offset_query_param, offset)

    def get_first_link(self):
        if self.count == 0:
            return None

        url = self.request.build_absolute_uri()
        return remove_query_param(url, self.offset_query_param)

    def get_paginated_response(self, data):
        return Response(
            {
                "results": data,
                "meta": {
                    "pagination": OrderedDict(
                        [
                            ("count", self.count),
                            ("limit", self.limit),
                            ("offset", self.offset),
                        ]
                    )
                },
                "links": OrderedDict(
                    [
                        ("first", self.get_first_link()),
                        ("last", self.get_last_link()),
                        ("next", self.get_next_link()),
                        ("prev", self.get_previous_link()),
                    ]
                ),
            }
        )



from .dev import *  # noqa

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = "example.urls_test"

JSON_API_FORMAT_FIELD_NAMES = "camelize"
JSON_API_FORMAT_TYPES = "camelize"
JSON_API_PLURALIZE_TYPES = True

REST_FRAMEWORK.update(  # noqa: F405
    {  # noqa
        "PAGE_SIZE": 1,
    }
)



import json
from io import BytesIO

import pytest
from rest_framework.exceptions import ParseError

from rest_framework_json_api.parsers import JSONParser
from rest_framework_json_api.utils import format_value
from tests.views import BasicModelViewSet


class TestJSONParser:
    @pytest.fixture
    def parser(self):
        return JSONParser()

    @pytest.fixture
    def parse(self, parser):
        def parse_wrapper(data, parser_context):
            stream = BytesIO(json.dumps(data).encode("utf-8"))
            return parser.parse(stream, None, parser_context)

        return parse_wrapper

    @pytest.fixture
    def parser_context(self, rf):
        return {"request": rf.post("/"), "kwargs": {}, "view": BasicModelViewSet()}

    @pytest.mark.parametrize(
        "format_field_names",
        [
            False,
            "dasherize",
            "camelize",
            "capitalize",
            "underscore",
        ],
    )
    def test_parse_formats_field_names(
        self,
        settings,
        format_field_names,
        parse,
        parser_context,
    ):
        settings.JSON_API_FORMAT_FIELD_NAMES = format_field_names

        data = {
            "data": {
                "id": "123",
                "type": "BasicModel",
                "attributes": {
                    format_value("test_attribute", format_field_names): "test-value"
                },
                "relationships": {
                    format_value("test_relationship", format_field_names): {
                        "data": {"type": "TestRelationship", "id": "123"}
                    }
                },
            }
        }

        result = parse(data, parser_context)
        assert result == {
            "id": "123",
            "test_attribute": "test-value",
            "test_relationship": {"id": "123", "type": "TestRelationship"},
        }

    def test_parse_extracts_meta(self, parse, parser_context):
        data = {
            "data": {
                "type": "BasicModel",
            },
            "meta": {"random_key": "random_value"},
        }

        result = parse(data, parser_context)
        assert result["_meta"] == data["meta"]

    def test_parse_with_default_arguments(self, parse):
        data = {
            "data": {
                "type": "BasicModel",
            },
        }
        result = parse(data, None)
        assert result == {}

    def test_parse_preserves_json_value_field_names(
        self, settings, parse, parser_context
    ):
        settings.JSON_API_FORMAT_FIELD_NAMES = "dasherize"

        data = {
            "data": {
                "type": "BasicModel",
                "attributes": {"json-value": {"JsonKey": "JsonValue"}},
            },
        }

        result = parse(data, parser_context)
        assert result["json_value"] == {"JsonKey": "JsonValue"}

    def test_parse_raises_error_on_empty_data(self, parse, parser_context):
        data = []

        with pytest.raises(ParseError) as excinfo:
            parse(data, parser_context)
        assert "Received document does not contain primary data" == str(excinfo.value)

    def test_parse_fails_on_list_of_objects(self, parse, parser_context):
        data = {
            "data": [
                {
                    "type": "BasicModel",
                    "attributes": {"json-value": {"JsonKey": "JsonValue"}},
                }
            ],
        }

        with pytest.raises(ParseError) as excinfo:
            parse(data, parser_context)

        assert (
            "Received data is not a valid JSON:API Resource Identifier Object"
            == str(excinfo.value)
        )

    def test_parse_fails_when_id_is_missing_on_patch(self, rf, parse, parser_context):
        parser_context["request"] = rf.patch("/")
        data = {
            "data": {
                "type": "BasicModel",
            },
        }

        with pytest.raises(ParseError) as excinfo:
            parse(data, parser_context)

        assert "The resource identifier object must contain an 'id' member" == str(
            excinfo.value
        )



from collections import OrderedDict

from django.db.models.fields import related
from django.utils.encoding import force_str
from rest_framework import serializers
from rest_framework.metadata import SimpleMetadata
from rest_framework.settings import api_settings
from rest_framework.utils.field_mapping import ClassLookupDict

from rest_framework_json_api.utils import format_field_name, get_related_resource_type


class JSONAPIMetadata(SimpleMetadata):
    """
    This is the JSON:API metadata implementation.
    It returns an ad-hoc set of information about the view.
    There are not any formalized standards for `OPTIONS` responses
    for us to base this on.
    """

    type_lookup = ClassLookupDict(
        {
            serializers.Field: "GenericField",
            serializers.RelatedField: "Relationship",
            serializers.BooleanField: "Boolean",
            serializers.NullBooleanField: "Boolean",
            serializers.CharField: "String",
            serializers.URLField: "URL",
            serializers.EmailField: "Email",
            serializers.RegexField: "Regex",
            serializers.SlugField: "Slug",
            serializers.IntegerField: "Integer",
            serializers.FloatField: "Float",
            serializers.DecimalField: "Decimal",
            serializers.DateField: "Date",
            serializers.DateTimeField: "DateTime",
            serializers.TimeField: "Time",
            serializers.ChoiceField: "Choice",
            serializers.MultipleChoiceField: "MultipleChoice",
            serializers.FileField: "File",
            serializers.ImageField: "Image",
            serializers.ListField: "List",
            serializers.DictField: "Dict",
            serializers.Serializer: "Serializer",
        }
    )

    try:
        relation_type_lookup = ClassLookupDict(
            {
                related.ManyToManyDescriptor: "ManyToMany",
                related.ReverseManyToOneDescriptor: "OneToMany",
                related.ForwardManyToOneDescriptor: "ManyToOne",
            }
        )
    except AttributeError:
        relation_type_lookup = ClassLookupDict(
            {
                related.ManyRelatedObjectsDescriptor: "ManyToMany",
                related.ReverseManyRelatedObjectsDescriptor: "ManyToMany",
                related.ForeignRelatedObjectsDescriptor: "OneToMany",
                related.ReverseSingleRelatedObjectDescriptor: "ManyToOne",
            }
        )

    def determine_metadata(self, request, view):
        metadata = OrderedDict()
        metadata["name"] = view.get_view_name()
        metadata["description"] = view.get_view_description()
        metadata["renders"] = [
            renderer.media_type for renderer in view.renderer_classes
        ]
        metadata["parses"] = [parser.media_type for parser in view.parser_classes]
        metadata["allowed_methods"] = view.allowed_methods
        if hasattr(view, "get_serializer"):
            actions = self.determine_actions(request, view)
            if actions:
                metadata["actions"] = actions
        return metadata

    def get_serializer_info(self, serializer):
        """
        Given an instance of a serializer, return a dictionary of metadata
        about its fields.
        """
        if hasattr(serializer, "child"):
            # If this is a `ListSerializer` then we want to examine the
            # underlying child serializer instance instead.
            serializer = serializer.child

        # Remove the URL field if present
        serializer.fields.pop(api_settings.URL_FIELD_NAME, None)

        return OrderedDict(
            [
                (format_field_name(field_name), self.get_field_info(field))
                for field_name, field in serializer.fields.items()
            ]
        )

    def get_field_info(self, field):
        """
        Given an instance of a serializer field, return a dictionary
        of metadata about it.
        """
        field_info = OrderedDict()
        serializer = field.parent

        if isinstance(field, serializers.ManyRelatedField):
            field_info["type"] = self.type_lookup[field.child_relation]
        else:
            field_info["type"] = self.type_lookup[field]

        try:
            serializer_model = getattr(serializer.Meta, "model")
            field_info["relationship_type"] = self.relation_type_lookup[
                getattr(serializer_model, field.field_name)
            ]
        except KeyError:
            pass
        except AttributeError:
            pass
        else:
            field_info["relationship_resource"] = get_related_resource_type(field)

        field_info["required"] = getattr(field, "required", False)

        attrs = [
            "read_only",
            "write_only",
            "label",
            "help_text",
            "min_length",
            "max_length",
            "min_value",
            "max_value",
            "initial",
        ]

        for attr in attrs:
            value = getattr(field, attr, None)
            if value is not None and value != "":
                field_info[attr] = force_str(value, strings_only=True)

        if getattr(field, "child", None):
            field_info["child"] = self.get_field_info(field.child)
        elif getattr(field, "fields", None):
            field_info["children"] = self.get_serializer_info(field)

        if (
            not field_info.get("read_only")
            and not field_info.get("relationship_resource")
            and hasattr(field, "choices")
        ):
            field_info["choices"] = [
                {
                    "value": choice_value,
                    "display_name": force_str(choice_name, strings_only=True),
                }
                for choice_value, choice_name in field.choices.items()
            ]

        if (
            hasattr(serializer, "included_serializers")
            and "relationship_resource" in field_info
        ):
            field_info["allows_include"] = (
                field.field_name in serializer.included_serializers
            )

        return field_info



from django.urls import reverse

from example.tests import TestBase


class GenericValidationTest(TestBase):
    """
    Test that a non serializer specific validation can be thrown and formatted
    """

    def setUp(self):
        super().setUp()
        self.url = reverse("user-validation", kwargs={"pk": self.miles.pk})

    def test_generic_validation_error(self):
        """
        Check error formatting
        """
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 400)

        expected = {
            "errors": [
                {
                    "status": "400",
                    "source": {"pointer": "/data"},
                    "detail": "Oh nohs!",
                    "code": "invalid",
                }
            ]
        }

        assert expected == response.json()



from datetime import datetime

import pytest
from django.urls import reverse

pytestmark = pytest.mark.django_db


def test_top_level_meta_for_list_view(blog, client):

    expected = {
        "data": [
            {
                "type": "blogs",
                "id": "1",
                "attributes": {"name": blog.name},
                "links": {"self": "http://testserver/blogs/1"},
                "relationships": {"tags": {"data": [], "meta": {"count": 0}}},
                "meta": {"copyright": datetime.now().year},
            }
        ],
        "links": {
            "first": "http://testserver/blogs?page%5Bnumber%5D=1",
            "last": "http://testserver/blogs?page%5Bnumber%5D=1",
            "next": None,
            "prev": None,
        },
        "meta": {
            "pagination": {"count": 1, "page": 1, "pages": 1},
            "apiDocs": "/docs/api/blogs",
        },
    }

    response = client.get(reverse("blog-list"))

    assert expected == response.json()


def test_top_level_meta_for_detail_view(blog, client):

    expected = {
        "data": {
            "type": "blogs",
            "id": "1",
            "attributes": {"name": blog.name},
            "relationships": {"tags": {"data": [], "meta": {"count": 0}}},
            "links": {"self": "http://testserver/blogs/1"},
            "meta": {"copyright": datetime.now().year},
        },
        "meta": {"apiDocs": "/docs/api/blogs"},
    }

    response = client.get(reverse("blog-detail", kwargs={"pk": blog.pk}))

    assert expected == response.json()



import json
from datetime import datetime

from django.test import RequestFactory, override_settings
from django.utils import timezone
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from rest_framework_json_api import serializers, views
from rest_framework_json_api.utils import format_resource_type

from example.factories import AuthorFactory, CommentFactory, EntryFactory
from example.models import Author, Blog, Comment, Entry
from example.serializers import (
    AuthorBioSerializer,
    AuthorTypeSerializer,
    EntrySerializer,
)
from example.tests import TestBase
from example.views import AuthorViewSet, BlogViewSet


class TestRelationshipView(APITestCase):
    def setUp(self):
        self.author = Author.objects.create(
            name="Super powerful superhero", email="i.am@lost.com"
        )
        self.blog = Blog.objects.create(name="Some Blog", tagline="It's a blog")
        self.other_blog = Blog.objects.create(
            name="Other blog", tagline="It's another blog"
        )
        self.first_entry = Entry.objects.create(
            blog=self.blog,
            headline="headline one",
            body_text="body_text two",
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=3,
        )
        self.second_entry = Entry.objects.create(
            blog=self.blog,
            headline="headline two",
            body_text="body_text one",
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=1,
        )
        self.first_comment = Comment.objects.create(
            entry=self.first_entry, body="This entry is cool", author=None
        )
        self.second_comment = Comment.objects.create(
            entry=self.second_entry, body="This entry is not cool", author=self.author
        )

    def test_get_entry_relationship_blog(self):
        url = reverse(
            "entry-relationships",
            kwargs={"pk": self.first_entry.id, "related_field": "blog"},
        )
        response = self.client.get(url)
        expected_data = {
            "type": format_resource_type("Blog"),
            "id": str(self.first_entry.blog.id),
        }

        assert response.data == expected_data

    def test_get_entry_relationship_invalid_field(self):
        response = self.client.get(
            "/entries/{}/relationships/invalid_field".format(self.first_entry.id)
        )

        assert response.status_code == 404

    def test_get_blog_relationship_entry_set(self):
        response = self.client.get(
            "/blogs/{}/relationships/entry_set".format(self.blog.id)
        )
        expected_data = [
            {"type": format_resource_type("Entry"), "id": str(self.first_entry.id)},
            {"type": format_resource_type("Entry"), "id": str(self.second_entry.id)},
        ]

        assert response.data == expected_data

    @override_settings(JSON_API_FORMAT_RELATED_LINKS="dasherize")
    def test_get_blog_relationship_entry_set_with_formatted_link(self):
        response = self.client.get(
            "/blogs/{}/relationships/entry-set".format(self.blog.id)
        )
        expected_data = [
            {"type": format_resource_type("Entry"), "id": str(self.first_entry.id)},
            {"type": format_resource_type("Entry"), "id": str(self.second_entry.id)},
        ]

        assert response.data == expected_data

    def test_put_entry_relationship_blog_returns_405(self):
        url = "/entries/{}/relationships/blog".format(self.first_entry.id)
        response = self.client.put(url, data={})
        assert response.status_code == 405

    def test_patch_invalid_entry_relationship_blog_returns_400(self):
        url = "/entries/{}/relationships/blog".format(self.first_entry.id)
        response = self.client.patch(url, data={"data": {"invalid": ""}})
        assert response.status_code == 400

    def test_relationship_view_errors_format(self):
        url = "/entries/{}/relationships/blog".format(self.first_entry.id)
        response = self.client.patch(url, data={"data": {"invalid": ""}})
        assert response.status_code == 400

        result = json.loads(response.content.decode("utf-8"))

        assert "data" not in result
        assert "errors" in result

    def test_get_empty_to_one_relationship(self):
        url = "/comments/{}/relationships/author".format(self.first_entry.id)
        response = self.client.get(url)
        expected_data = None

        assert response.data == expected_data

    def test_get_to_many_relationship_self_link(self):
        url = "/authors/{}/relationships/comments".format(self.author.id)

        response = self.client.get(url)
        expected_data = {
            "links": {"self": "http://testserver/authors/1/relationships/comments"},
            "data": [
                {
                    "id": str(self.second_comment.id),
                    "type": format_resource_type("Comment"),
                }
            ],
        }
        assert json.loads(response.content.decode("utf-8")) == expected_data

    def test_patch_to_one_relationship(self):
        url = "/entries/{}/relationships/blog".format(self.first_entry.id)
        request_data = {
            "data": {
                "type": format_resource_type("Blog"),
                "id": str(self.other_blog.id),
            }
        }
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data == request_data["data"]

        response = self.client.get(url)
        assert response.data == request_data["data"]

    def test_patch_one_to_many_relationship(self):
        url = "/blogs/{}/relationships/entry_set".format(self.first_entry.id)
        request_data = {
            "data": [
                {"type": format_resource_type("Entry"), "id": str(self.first_entry.id)},
            ]
        }
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data == request_data["data"]

        response = self.client.get(url)
        assert response.data == request_data["data"]

        # retry a second time should end up with same result
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data == request_data["data"]

        response = self.client.get(url)
        assert response.data == request_data["data"]

    def test_patch_one_to_many_relaitonship_with_none(self):
        url = "/blogs/{}/relationships/entry_set".format(self.first_entry.id)
        request_data = {"data": None}
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data == []

        response = self.client.get(url)
        assert response.data == []

    def test_patch_many_to_many_relationship(self):
        url = "/entries/{}/relationships/authors".format(self.first_entry.id)
        request_data = {
            "data": [
                {"type": format_resource_type("Author"), "id": str(self.author.id)},
            ]
        }
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data == request_data["data"]

        response = self.client.get(url)
        assert response.data == request_data["data"]

        # retry a second time should end up with same result
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data == request_data["data"]

        response = self.client.get(url)
        assert response.data == request_data["data"]

    def test_post_to_one_relationship_should_fail(self):
        url = "/entries/{}/relationships/blog".format(self.first_entry.id)
        request_data = {
            "data": {
                "type": format_resource_type("Blog"),
                "id": str(self.other_blog.id),
            }
        }
        response = self.client.post(url, data=request_data)
        assert response.status_code == 405, response.content.decode()

    def test_post_to_many_relationship_with_no_change(self):
        url = "/entries/{}/relationships/comments".format(self.first_entry.id)
        request_data = {
            "data": [
                {
                    "type": format_resource_type("Comment"),
                    "id": str(self.first_comment.id),
                },
            ]
        }
        response = self.client.post(url, data=request_data)
        assert response.status_code == 204, response.content.decode()
        assert len(response.rendered_content) == 0, response.rendered_content.decode()

    def test_post_to_many_relationship_with_change(self):
        url = "/entries/{}/relationships/comments".format(self.first_entry.id)
        request_data = {
            "data": [
                {
                    "type": format_resource_type("Comment"),
                    "id": str(self.second_comment.id),
                },
            ]
        }
        response = self.client.post(url, data=request_data)
        assert response.status_code == 200, response.content.decode()

        assert request_data["data"][0] in response.data

    def test_delete_to_one_relationship_should_fail(self):
        url = "/entries/{}/relationships/blog".format(self.first_entry.id)
        request_data = {
            "data": {
                "type": format_resource_type("Blog"),
                "id": str(self.other_blog.id),
            }
        }
        response = self.client.delete(url, data=request_data)
        assert response.status_code == 405, response.content.decode()

    def test_delete_relationship_overriding_with_none(self):
        url = "/comments/{}".format(self.second_comment.id)
        request_data = {
            "data": {
                "type": "comments",
                "id": self.second_comment.id,
                "relationships": {"author": {"data": None}},
            }
        }
        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200, response.content.decode()
        assert response.data["author"] is None

    def test_delete_to_many_relationship_with_no_change(self):
        url = "/entries/{}/relationships/comments".format(self.first_entry.id)
        request_data = {
            "data": [
                {
                    "type": format_resource_type("Comment"),
                    "id": str(self.second_comment.id),
                },
            ]
        }
        response = self.client.delete(url, data=request_data)
        assert response.status_code == 204, response.content.decode()
        assert len(response.rendered_content) == 0, response.rendered_content.decode()

    def test_delete_one_to_many_relationship_with_not_null_constraint(self):
        url = "/entries/{}/relationships/comments".format(self.first_entry.id)
        request_data = {
            "data": [
                {
                    "type": format_resource_type("Comment"),
                    "id": str(self.first_comment.id),
                },
            ]
        }
        response = self.client.delete(url, data=request_data)
        assert response.status_code == 409, response.content.decode()

    def test_delete_to_many_relationship_with_change(self):
        url = "/authors/{}/relationships/comments".format(self.author.id)
        request_data = {
            "data": [
                {
                    "type": format_resource_type("Comment"),
                    "id": str(self.second_comment.id),
                },
            ]
        }
        response = self.client.delete(url, data=request_data)
        assert response.status_code == 200, response.content.decode()

    def test_new_comment_data_patch_to_many_relationship(self):
        entry = EntryFactory(blog=self.blog, authors=(self.author,))
        comment = CommentFactory(entry=entry)

        url = "/authors/{}/relationships/comments".format(self.author.id)
        request_data = {
            "data": [
                {"type": format_resource_type("Comment"), "id": str(comment.id)},
            ]
        }
        previous_response = {
            "data": [{"type": "comments", "id": str(self.second_comment.id)}],
            "links": {
                "self": "http://testserver/authors/{}/relationships/comments".format(
                    self.author.id
                )
            },
        }

        response = self.client.get(url)
        assert response.status_code == 200
        assert response.json() == previous_response

        new_patched_response = {
            "data": [{"type": "comments", "id": str(comment.id)}],
            "links": {
                "self": "http://testserver/authors/{}/relationships/comments".format(
                    self.author.id
                )
            },
        }

        response = self.client.patch(url, data=request_data)
        assert response.status_code == 200
        assert response.json() == new_patched_response

        assert Comment.objects.filter(id=self.second_comment.id).exists()

    def test_options_entry_relationship_blog(self):
        url = reverse(
            "entry-relationships",
            kwargs={"pk": self.first_entry.id, "related_field": "blog"},
        )
        response = self.client.options(url)
        expected_data = {
            "data": {
                "name": "Entry Relationship",
                "description": "",
                "renders": ["application/vnd.api+json", "text/html"],
                "parses": [
                    "application/vnd.api+json",
                    "application/x-www-form-urlencoded",
                    "multipart/form-data",
                ],
                "allowed_methods": [
                    "GET",
                    "POST",
                    "PATCH",
                    "DELETE",
                    "HEAD",
                    "OPTIONS",
                ],
                "actions": {"POST": {}},
            }
        }
        assert response.json() == expected_data


class TestRelatedMixin(APITestCase):
    def setUp(self):
        self.author = AuthorFactory()

    def _get_view(self, kwargs):
        factory = APIRequestFactory()
        request = Request(factory.get("", content_type="application/vnd.api+json"))
        return AuthorViewSet(request=request, kwargs=kwargs)

    def test_get_related_field_name(self):
        kwargs = {"pk": self.author.id, "related_field": "bio"}
        view = self._get_view(kwargs)
        got = view.get_related_field_name()
        self.assertEqual(got, kwargs["related_field"])

    def test_get_related_instance_serializer_field(self):
        kwargs = {"pk": self.author.id, "related_field": "bio"}
        view = self._get_view(kwargs)
        got = view.get_related_instance()
        self.assertEqual(got, self.author.bio)

    def test_get_related_instance_model_field(self):
        kwargs = {"pk": self.author.id, "related_field": "id"}
        view = self._get_view(kwargs)
        got = view.get_related_instance()
        self.assertEqual(got, self.author.id)

    def test_get_related_serializer_class(self):
        kwargs = {"pk": self.author.id, "related_field": "bio"}
        view = self._get_view(kwargs)
        got = view.get_related_serializer_class()
        self.assertEqual(got, AuthorBioSerializer)

    def test_get_related_serializer_class_many(self):
        kwargs = {"pk": self.author.id, "related_field": "entries"}
        view = self._get_view(kwargs)
        got = view.get_related_serializer_class()
        self.assertEqual(got, EntrySerializer)

    def test_get_serializer_comes_from_included_serializers(self):
        kwargs = {"pk": self.author.id, "related_field": "type"}
        view = self._get_view(kwargs)
        related_serializers = view.get_serializer_class().related_serializers
        delattr(view.get_serializer_class(), "related_serializers")
        got = view.get_related_serializer_class()
        self.assertEqual(got, AuthorTypeSerializer)
        view.get_serializer_class().related_serializers = related_serializers

    def test_get_related_serializer_class_raises_error(self):
        kwargs = {"pk": self.author.id, "related_field": "unknown"}
        view = self._get_view(kwargs)
        self.assertRaises(NotFound, view.get_related_serializer_class)

    def test_retrieve_related_single_reverse_lookup(self):
        url = reverse(
            "author-related", kwargs={"pk": self.author.pk, "related_field": "bio"}
        )
        resp = self.client.get(url, data={"include": "metadata"})
        expected = {
            "data": {
                "type": "authorBios",
                "id": str(self.author.bio.id),
                "relationships": {
                    "author": {"data": {"type": "authors", "id": str(self.author.id)}},
                    "metadata": {
                        "data": {
                            "id": str(self.author.bio.metadata.id),
                            "type": "authorBioMetadata",
                        }
                    },
                },
                "attributes": {"body": str(self.author.bio.body)},
            },
            "included": [
                {
                    "attributes": {"body": str(self.author.bio.metadata.body)},
                    "id": str(self.author.bio.metadata.id),
                    "type": "authorBioMetadata",
                }
            ],
        }
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), expected)

    def test_retrieve_related_single(self):
        url = reverse(
            "author-related",
            kwargs={"pk": self.author.type.pk, "related_field": "type"},
        )
        resp = self.client.get(url)
        expected = {
            "data": {
                "type": "authorTypes",
                "id": str(self.author.type.id),
                "attributes": {"name": str(self.author.type.name)},
            }
        }
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), expected)

    def test_retrieve_related_many(self):
        entry = EntryFactory(authors=self.author)
        url = reverse(
            "author-related", kwargs={"pk": self.author.pk, "related_field": "entries"}
        )
        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTrue(isinstance(resp.json()["data"], list))
        self.assertEqual(len(resp.json()["data"]), 1)
        self.assertEqual(resp.json()["data"][0]["id"], str(entry.id))

    def test_retrieve_related_many_hyperlinked(self):
        comment = CommentFactory(author=self.author)
        url = reverse(
            "author-related", kwargs={"pk": self.author.pk, "related_field": "comments"}
        )
        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTrue(isinstance(resp.json()["data"], list))
        self.assertEqual(len(resp.json()["data"]), 1)
        self.assertEqual(resp.json()["data"][0]["id"], str(comment.id))

    def test_retrieve_related_None(self):
        kwargs = {"pk": self.author.pk, "related_field": "first_entry"}
        url = reverse("author-related", kwargs=kwargs)
        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), {"data": None})

    @override_settings(JSON_API_FORMAT_RELATED_LINKS="dasherize")
    def test_retrieve_related_with_formatted_link(self):
        first_entry = EntryFactory(authors=(self.author,))

        kwargs = {"pk": self.author.pk, "related_field": "first-entry"}
        url = reverse("author-related", kwargs=kwargs)
        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()["data"]["id"], str(first_entry.id))


class TestValidationErrorResponses(TestBase):
    def test_if_returns_error_on_empty_post(self):
        view = BlogViewSet.as_view({"post": "create"})
        response = self._get_create_response("{}", view)
        self.assertEqual(400, response.status_code)
        expected = [
            {
                "detail": "Received document does not contain primary data",
                "status": "400",
                "source": {"pointer": "/data"},
                "code": "parse_error",
            }
        ]
        self.assertEqual(expected, response.data)

    def test_if_returns_error_on_missing_form_data_post(self):
        view = BlogViewSet.as_view({"post": "create"})
        response = self._get_create_response(
            '{"data":{"attributes":{},"type":"blogs"}}', view
        )
        self.assertEqual(400, response.status_code)
        expected = [
            {
                "status": "400",
                "detail": "This field is required.",
                "source": {"pointer": "/data/attributes/name"},
                "code": "required",
            }
        ]
        self.assertEqual(expected, response.data)

    def test_if_returns_error_on_bad_endpoint_name(self):
        view = BlogViewSet.as_view({"post": "create"})
        response = self._get_create_response(
            '{"data":{"attributes":{},"type":"bad"}}', view
        )
        self.assertEqual(409, response.status_code)
        expected = [
            {
                "detail": (
                    "The resource object's type (bad) is not the type that constitute the collection "
                    "represented by the endpoint (blogs)."
                ),
                "source": {"pointer": "/data"},
                "status": "409",
                "code": "error",
            }
        ]
        self.assertEqual(expected, response.data)

    def _get_create_response(self, data, view):
        factory = RequestFactory()
        request = factory.post("/", data, content_type="application/vnd.api+json")
        user = self.create_user("user", "pass")
        force_authenticate(request, user)
        return view(request)


class TestModelViewSet(TestBase):
    def setUp(self):
        self.author = Author.objects.create(
            name="Super powerful superhero", email="i.am@lost.com"
        )
        self.blog = Blog.objects.create(name="Some Blog", tagline="It's a blog")

    def test_no_content_response(self):
        url = "/blogs/{}".format(self.blog.pk)
        response = self.client.delete(url)
        assert response.status_code == 204, response.rendered_content.decode()
        assert len(response.rendered_content) == 0, response.rendered_content.decode()


class TestBlogViewSet(APITestCase):
    def setUp(self):
        self.blog = Blog.objects.create(name="Some Blog", tagline="It's a blog")
        self.entry = Entry.objects.create(
            blog=self.blog,
            headline="headline one",
            body_text="body_text two",
        )

    def test_get_object_gives_correct_blog(self):
        url = reverse("entry-blog", kwargs={"entry_pk": self.entry.id})
        resp = self.client.get(url)
        expected = {
            "data": {
                "attributes": {"name": self.blog.name},
                "id": "{}".format(self.blog.id),
                "links": {"self": "http://testserver/blogs/{}".format(self.blog.id)},
                "meta": {"copyright": datetime.now().year},
                "relationships": {"tags": {"data": [], "meta": {"count": 0}}},
                "type": "blogs",
            },
            "meta": {"apiDocs": "/docs/api/blogs"},
        }
        got = resp.json()
        self.assertEqual(got, expected)


class TestEntryViewSet(APITestCase):
    def setUp(self):
        self.blog = Blog.objects.create(name="Some Blog", tagline="It's a blog")
        self.first_entry = Entry.objects.create(
            blog=self.blog,
            headline="headline two",
            body_text="body_text two",
        )
        self.second_entry = Entry.objects.create(
            blog=self.blog,
            headline="headline two",
            body_text="body_text two",
        )
        self.maxDiff = None

    def test_get_object_gives_correct_entry(self):
        url = reverse("entry-featured", kwargs={"entry_pk": self.first_entry.id})
        resp = self.client.get(url)
        expected = {
            "data": {
                "attributes": {
                    "bodyText": self.second_entry.body_text,
                    "headline": self.second_entry.headline,
                    "modDate": self.second_entry.mod_date,
                    "pubDate": self.second_entry.pub_date,
                },
                "id": "{}".format(self.second_entry.id),
                "meta": {"bodyFormat": "text"},
                "relationships": {
                    "authors": {"data": [], "meta": {"count": 0}},
                    "blog": {
                        "data": {
                            "id": "{}".format(self.second_entry.blog_id),
                            "type": "blogs",
                        }
                    },
                    "blogHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/{}"
                            "/blog".format(self.second_entry.id),
                            "self": "http://testserver/entries/{}"
                            "/relationships/blog_hyperlinked".format(
                                self.second_entry.id
                            ),
                        }
                    },
                    "comments": {"data": [], "meta": {"count": 0}},
                    "commentsHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/{}"
                            "/comments".format(self.second_entry.id),
                            "self": "http://testserver/entries/{}/relationships"
                            "/comments_hyperlinked".format(self.second_entry.id),
                        }
                    },
                    "featuredHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/{}"
                            "/featured".format(self.second_entry.id),
                            "self": "http://testserver/entries/{}/relationships"
                            "/featured_hyperlinked".format(self.second_entry.id),
                        }
                    },
                    "suggested": {
                        "data": [{"id": "1", "type": "entries"}],
                        "links": {
                            "related": "http://testserver/entries/{}"
                            "/suggested/".format(self.second_entry.id),
                            "self": "http://testserver/entries/{}"
                            "/relationships/suggested".format(self.second_entry.id),
                        },
                    },
                    "suggestedHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/{}"
                            "/suggested/".format(self.second_entry.id),
                            "self": "http://testserver/entries/{}/relationships"
                            "/suggested_hyperlinked".format(self.second_entry.id),
                        }
                    },
                    "tags": {"data": [], "meta": {"count": 0}},
                },
                "type": "posts",
            }
        }
        got = resp.json()
        self.assertEqual(got, expected)


class BasicAuthorSerializer(serializers.ModelSerializer):
    class Meta:
        model = Author
        fields = ("name",)


class ReadOnlyViewSetWithCustomActions(views.ReadOnlyModelViewSet):
    queryset = Author.objects.all()
    serializer_class = BasicAuthorSerializer

    @action(detail=False, methods=["get", "post", "patch", "delete"])
    def group_action(self, request):
        return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=True, methods=["get", "post", "patch", "delete"])
    def item_action(self, request, pk):
        return Response(status=status.HTTP_204_NO_CONTENT)


class TestReadonlyModelViewSet(TestBase):
    """
    Test if ReadOnlyModelViewSet allows to have custom actions with POST, PATCH, DELETE methods
    """

    factory = RequestFactory()
    viewset_class = ReadOnlyViewSetWithCustomActions
    media_type = "application/vnd.api+json"

    def test_group_action_allows_get(self):
        view = self.viewset_class.as_view({"get": "group_action"})
        request = self.factory.get("/")
        response = view(request)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_group_action_allows_post(self):
        view = self.viewset_class.as_view({"post": "group_action"})
        request = self.factory.post("/", "{}", content_type=self.media_type)
        response = view(request)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_group_action_allows_patch(self):
        view = self.viewset_class.as_view({"patch": "group_action"})
        request = self.factory.patch("/", "{}", content_type=self.media_type)
        response = view(request)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_group_action_allows_delete(self):
        view = self.viewset_class.as_view({"delete": "group_action"})
        request = self.factory.delete("/", "{}", content_type=self.media_type)
        response = view(request)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_item_action_allows_get(self):
        view = self.viewset_class.as_view({"get": "item_action"})
        request = self.factory.get("/")
        response = view(request, pk="1")
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_item_action_allows_post(self):
        view = self.viewset_class.as_view({"post": "item_action"})
        request = self.factory.post("/", "{}", content_type=self.media_type)
        response = view(request, pk="1")
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_item_action_allows_patch(self):
        view = self.viewset_class.as_view({"patch": "item_action"})
        request = self.factory.patch("/", "{}", content_type=self.media_type)
        response = view(request, pk="1")
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_item_action_allows_delete(self):
        view = self.viewset_class.as_view({"delete": "item_action"})
        request = self.factory.delete("/", "{}", content_type=self.media_type)
        response = view(request, pk="1")
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)



from rest_framework import filters as drf_filters

from rest_framework_json_api import filters as dja_filters
from rest_framework_json_api.django_filters import backends

from example.views import EntryViewSet


class DummyEntryViewSet(EntryViewSet):
    filter_backends = (
        dja_filters.QueryParameterValidationFilter,
        dja_filters.OrderingFilter,
        backends.DjangoFilterBackend,
        drf_filters.SearchFilter,
    )
    filterset_fields = {
        "id": ("exact",),
        "headline": ("exact", "contains"),
        "blog__name": ("contains",),
    }

    def __init__(self, **kwargs):
        # dummy up self.request since PreloadIncludesMixin expects it to be defined
        self.request = None
        super().__init__(**kwargs)


def test_filters_get_schema_params():
    """
    test all my filters for `get_schema_operation_parameters()`
    """
    # list of tuples: (filter, expected result)
    filters = [
        (dja_filters.QueryParameterValidationFilter, []),
        (
            backends.DjangoFilterBackend,
            [
                {
                    "name": "filter[id]",
                    "required": False,
                    "in": "query",
                    "description": "id",
                    "schema": {"type": "string"},
                },
                {
                    "name": "filter[headline]",
                    "required": False,
                    "in": "query",
                    "description": "headline",
                    "schema": {"type": "string"},
                },
                {
                    "name": "filter[headline.contains]",
                    "required": False,
                    "in": "query",
                    "description": "headline__contains",
                    "schema": {"type": "string"},
                },
                {
                    "name": "filter[blog.name.contains]",
                    "required": False,
                    "in": "query",
                    "description": "blog__name__contains",
                    "schema": {"type": "string"},
                },
            ],
        ),
        (
            dja_filters.OrderingFilter,
            [
                {
                    "name": "sort",
                    "required": False,
                    "in": "query",
                    "description": "Which field to use when ordering the results.",
                    "schema": {"type": "string"},
                }
            ],
        ),
        (
            drf_filters.SearchFilter,
            [
                {
                    "name": "filter[search]",
                    "required": False,
                    "in": "query",
                    "description": "A search term.",
                    "schema": {"type": "string"},
                }
            ],
        ),
    ]
    view = DummyEntryViewSet()

    for c, expected in filters:
        f = c()
        result = f.get_schema_operation_parameters(view)
        assert len(result) == len(expected)
        if len(result) == 0:
            continue
        # py35: the result list/dict ordering isn't guaranteed
        for res_item in result:
            assert "name" in res_item
            for exp_item in expected:
                if res_item["name"] == exp_item["name"]:
                    assert res_item == exp_item
